import numpy as np
import time 

import parakeet

from parakeet import jit, pmap2 
from parakeet.testing_helpers import eq, run_local_tests


def winmin(x):
  m,n = x.shape
  v = x[0,0]
  for i in range(m):
    for j in range(n):
      v2 = x[i,j]
      if v2 < v:
        v = v2
  return v

def winmax(x):
  m,n = x.shape
  v = x[0,0]
  for i in range(m):
    for j in range(n):
      v2 = x[i,j]
      if v2 > v:
        v = v2
  return v

def erode(X, window_size = (3,3)):
  return pmap2(winmin, X, window_size)
 
def dilate(X, window_size = (3,3)):
  return pmap2(winmax, X, window_size)



X = np.array([[0, 0,   0,   0,   0],
              [0, 0.1,  0.2, 0.3, 0],
              [0, 0.3,  0.4, 0.3, 0],
              [0, 0.2,  0.1, 0.5, 0],
              [0, 0,    0,   0,   0]])

# what we would expect after a 3x3 erosion 
X_erode = np.array([[0, 0, 0,   0, 0],
                    [0, 0, 0,   0, 0],
                    [0, 0, 0.1, 0, 0],
                    [0, 0, 0,   0, 0],
                    [0, 0, 0,   0, 0]])

X_dilate = np.array([[0.1, 0.2, 0.3, 0.3, 0.3],
                     [0.3, 0.4, 0.4, 0.4, 0.3],
                     [0.3, 0.4, 0.5, 0.5, 0.5],
                     [0.3, 0.4, 0.5, 0.5, 0.5],
                     [0.2, 0.2, 0.5, 0.5, 0.5]])

def test_erode():
  res = erode(X, (3,3))
  assert res.shape == X_erode.shape, "Expected shape %s but got %s" % (X_erode.shape, res.shape)
  assert (res == X_erode).all(), "Expected %s but got %s" % (X_erode, res)

def test_dilate():
  res = dilate(X, (3,3))
  assert res.shape == X_dilate.shape, "Expected shape %s but got %s" % (X_dilate.shape, res.shape)
  assert (res == X_dilate).all(), "Expected %s but got %s" % (X_dilate, res)



def dilate_1d_naive(x_strip, y_strip, window_size):
  """
  Given a 1-dimensional input and 1-dimensional output, 
  fill output with 1d dilation of input 
  """
  nelts = len(x_strip)
  half = window_size / 2 
  for i in xrange(nelts):
    left_idx = max(i-half,0)
    right_idx = min(i+half+1, nelts)
    currmax = x_strip[left_idx]
    for j in xrange(left_idx+1, right_idx):
      elt = x_strip[j]
      if elt > currmax:
        currmax = elt
    y_strip[i] = currmax 

@jit 
def dilate_decompose(x, window_size): 
  m,n = x.shape
  k,l = window_size
  y = np.empty_like(x)
  z = np.empty_like(x)
  for row_idx in xrange(m):
    dilate_1d_naive(x[row_idx,:], y[row_idx,:], k)
  for col_idx in xrange(n):
    dilate_1d_naive(y[:, col_idx], z[:, col_idx], l)
  return z

def test_dilate_decompose():
  res = dilate_decompose(X, (3,3))
  assert res.shape == X_dilate.shape, "Expected shape %s but got %s" % (X_dilate.shape, res.shape)
  assert (res == X_dilate).all(), "Original \n%s Expected dilation \n%s but got \n%s unequal elts \n%s" % \
    (X, X_dilate, res, X_dilate != res)

def morph_open(x, erode_shape, dilate_shape = None):
  if dilate_shape is None: 
    dilate_shape = erode_shape
  return dilate(erode(x, erode_shape), dilate_shape)


def morph_close(x, dilate_shape, erode_shape = None):
  if erode_shape is None: 
    erode_shape = dilate_shape
  return erode(dilate(x, dilate_shape), erode_shape)


plot_rgb = False 

if __name__ == '__main__':
  run_local_tests() 



from .. import syntax
from ..ndtypes import ArrayT, SliceT, ScalarT, TupleT, Int64, StructT, ClosureT, IntT, NoneT 
from ..syntax import Attribute, TupleProj, Var, ClosureElt
from ..shape_inference import shape_env, shape
from ..transforms import Transform


class Counter(object):
  """
  Auto-incrementing counter
  """
  def __init__(self):
    self.n = 0
    
  def get(self):
    n = self.n 
    self.n += 1
    return n 
  
class ShapeElimination(Transform):
  
   
  def pre_apply(self, fn):
    self.shape_env = shape_env(fn)
    # map from shape vars to expressions 
    self.shape_vars = {}
    self.shape_var_counter = Counter()
    input_vars = [Var(arg_name, type = arg_type)
                  for arg_name, arg_type in zip(fn.arg_names, fn.input_types)]
    self.fill_shape_vars_list(input_vars)
    
    
      
  def fill_shape_vars(self, expr):
    t = expr.type 
    c = t.__class__ 
    if c is TupleT:
      elts = [TupleProj(expr,i,type=elt_t) 
              for i, elt_t in enumerate(t.elt_types)]
      self.fill_shape_vars_list(elts)
    elif c is ClosureT:
      elts = [ClosureElt(expr,i,type=elt_t) 
              for i, elt_t in enumerate(t.arg_types)]
      self.fill_shape_vars_list(elts)
    elif c is ArrayT:
      shape = Attribute(expr, 'shape', t.shape_t)
      dims = [TupleProj(shape,i,type=Int64) for i in xrange(t.rank)]
      self.fill_shape_vars_list(dims)
    elif c is StructT:
      fields = [Attribute(expr, field_name, type = field_type)
                for field_name, field_type in zip(t.field_names, t.field_types)]
      self.fill_shape_vars_list(fields)
    elif c is SliceT:
      fields = []
      start_t = expr.type.start_type 
      if start_t != NoneT: fields.append(Attribute(expr, 'start', type = start_t))
      stop_t = expr.type.stop_type 
      if stop_t != NoneT: fields.append(Attribute(expr, 'stop', type = stop_t))
      step_t = expr.type.step_type 
      if step_t != NoneT: fields.append(Attribute(expr, 'step', type = step_t))
      self.fill_shape_vars_list(fields)
    elif isinstance(t, ScalarT):
      self.shape_vars[self.shape_var_counter.get()] = expr 
      
      
  def fill_shape_vars_list(self, arg_exprs):
    for e in arg_exprs:
      self.fill_shape_vars(e)

  def transform_Var(self, expr):
    #print expr, self.shape_env.get(expr.name) 
    if expr.name in self.shape_env:
      v = self.shape_env[expr.name]
      # don't trust shape propagation to correctly handle floating point errors
      if v.__class__ is shape.Const and isinstance(expr.type, IntT):
        return syntax.Const(value = v.value, type = expr.type)
      elif v.__class__ is shape.Var: 
        new_expr = self.shape_vars[v.num]
        
        if new_expr != expr:
          return self.cast(self.shape_vars[v.num], expr.type)
    return expr
  
  def transform_lhs(self, lhs):
    return lhs


"""
Example taken from Numba discussion list 
https://groups.google.com/a/continuum.io/forum/#!topic/numba-users/pCpNGdtKlRc
"""

import math 
import numpy as np 
from parakeet import testing_helpers

def norm(vec):
    """ Calculate the norm of a 3d vector. """
    return math.sqrt(vec[0]**2 + vec[1]**2 + vec[2]**2)

def normalize(vec):
    """ Calculate the normalized vector (norm: one). """
    return vec / norm(vec)


def test_normalize():
  v = np.array((1.0, 2.0, 3.0))
  testing_helpers.expect(normalize, [v], normalize(v))

if __name__ == "__main__":
  testing_helpers.run_local_tests()




import numpy as np
from parakeet import jit  
from parakeet.testing_helpers import expect_eq, run_local_tests

@jit 
def add(x,y):
  return x + y

def test_2d_scalar():
  x = np.zeros((2,3))
  res = add(x, 1)
  expect_eq(res, x + 1)

def test_2d_1d():
  x = np.zeros((2,3))
  y = np.ones((3,))
  res = add(x, y)
  expect_eq(res, x + y)

def test_2d_2d():
  x = np.zeros((2,3))
  y = np.ones((2,3))
  res = add(x, y)
  expect_eq(res, x + y)
  


if __name__ == '__main__':
  run_local_tests()
  





import time 
import numpy as np 
from numpy import sqrt, int32, mod, double 
 
def calculate_distance(template, dx, dy) : 
    side_length = template.shape[0]
    # where is the center position
    cen = side_length/2
    
    for i in range(side_length) : 
        for j in range(side_length) : 
            template[i,j] = sqrt(((i-cen)*dx)**2 + ((j-cen)*dy)**2)




def kernel_loops(ds, h):
  fs = np.zeros_like(ds)
  denom = (np.pi*h**3)
  for i in xrange(ds.shape[0]):
    for j in xrange(ds.shape[1]):
      d = ds[i,j] 
      if d < 1 : 
        fs[i,j] = (1.-(3./2)*d**2 + (3./4.)*d**3) / denom 
      elif d <= 2.0 :
        fs[i,j] = 0.25*(2.-d)**3 / denom  
  return fs 
   


def physical_to_pixel(xpos,xmin,dx) : 
    return int32((xpos-xmin)/dx)

"""

def template_render_image(s,nx,ny,xmin,xmax,ymin,ymax,qty='rho',timing = False,two_d=0):
    
    time_init = time.clock()
    
    xs,ys,zs,hs,qts,mass,rhos = [s[arr] for arr in ['x','y','z','smooth',qty,'mass','rho']]

    # ----------------------
    # setup the global image
    # ----------------------
    image = np.zeros((nx,ny))
    
    dx = (xmax-xmin)/nx
    dy = (ymax-ymin)/ny
    
    x_start = xmin+dx/2
    y_start = ymin+dy/2

    zplane = 0.0

    # ------------------------------------
    # trim particles based on image limits
    # ------------------------------------
    start = time.clock()
    ind = np.where((xs + 2*hs > xmin) & (xs - 2*hs < xmax) & 
                   (ys + 2*hs > ymin) & (ys - 2*hs < ymax) &
                   (np.abs(zs-zplane)*(1-two_d) < 2*hs))[0]

    xs,ys,zs,hs,qts,mass,rhos = (xs[ind],ys[ind],zs[ind],hs[ind],qts[ind],mass[ind],rhos[ind])
    if timing: print '<<< Initial particle selection took %f s'%(time.clock()-start)

    # set the render quantity 
    qts *= mass/rhos

    #
    # bin particles by how many pixels they need in their kernel
    #
    start = time.clock()
    npix = 2.0*hs/dx
    dbin = np.digitize(npix,np.arange(1,npix.max()))
    dbin_sortind = dbin.argsort()
    dbin_sorted = dbin[dbin_sortind]
    xs,ys,zs,hs,qts = (xs[dbin_sortind],ys[dbin_sortind],zs[dbin_sortind],hs[dbin_sortind],qts[dbin_sortind])
    if timing: print '<<< Bin sort done in %f'%(time.clock()-start)

    # ---------------------
    # process the particles 
    # ---------------------
    start = time.clock()
    image = template_kernel_cpu(xs,ys,qts,hs,nx,ny,xmin,xmax,ymin,ymax,two_d)
    if timing: print '<<< Rendering %d particles took %f s'%(len(xs),
                                                             time.clock()-start)
    
    if timing: print '<<< Total time: %f s'%(time.clock()-time_init)

    return image

"""

def template_kernel_cpu(xs,ys,qts,hs,nx,ny,xmin,xmax,ymin,ymax,two_d) : 
    # ------------------
    # create local image 
    # ------------------
    image = np.zeros((nx,ny),dtype=np.float)

    Npart = len(hs) 
    dx = (xmax-xmin)/nx
    dy = (ymax-ymin)/ny

       
    # ------------------------------
    # start the loop through kernels
    # ------------------------------
    kmax = int(np.ceil(hs.max()*2.0/dx*2.0))
    kmin = int(np.floor(hs.min()*2.0/dx*2.0))
    # make sure kmin and kmax are odd
    
    kmax += (1 - mod(kmax,2))
    kmin += (1 - mod(kmin,2))
    
    kmin = max(1,kmin)
    
    kernel_base = np.ones((kmax,kmax))
    kernel = np.ones((kmax,kmax))
    calculate_distance(kernel_base,dx,dy)
    
    max_d_curr = 0.0
    start_ind = 0
    end_ind = 0
    for k in xrange(kmin,kmax+2,2) : 
        # ---------------------------------
        # the max. distance for this kernel
        # ---------------------------------
        max_d_curr = dx*np.floor(k/2.0)
        if max_d_curr < dx/2.0 : 
          max_d_curr = dx/2.0

        i_max_d = double(1./max_d_curr)
        # -------------------------------------------------
        # find the chunk of particles that need this kernel
        # -------------------------------------------------
        end_ind = 0 
        for i in xrange(start_ind,Npart): 
            if 2*hs[end_ind] < max_d_curr:
              end_ind = i
            
        
        Nper_kernel = end_ind-start_ind
        
        # -------------------------------------------------------------------------
        # only continue with kernel generation if there are particles that need it!
        # -------------------------------------------------------------------------
        if Nper_kernel > 0 : 
            kernel = kernel_base[kmax/2-k/2:kmax/2+k/2+1,
                                 kmax/2-k/2:kmax/2+k/2+1]
            kernel = kernel_loops(kernel*i_max_d*2.0,1.0)
            kernel *= 8*i_max_d*i_max_d*i_max_d # kernel / h**3
        
            # --------------------------------
            # paint each particle on the image
            # --------------------------------
            for pind in xrange(start_ind,end_ind) : 
                x,y,h,qt = [xs[pind],ys[pind],hs[pind],qts[pind]]
                
                # set the minimum h to be equal to half pixel width
                #                h = max_d_curr*.5
                #h = max(h,0.55*dx)
                
                # particle pixel center
                xpos = physical_to_pixel(x,xmin,dx)
                ypos = physical_to_pixel(y,ymin,dy)
    
                left  = xpos-k/2
                upper = ypos-k/2

                for i in xrange(0,k) : 
                    for j in xrange(0,k): 
                        if ((i+left>=0) and (i+left < nx) and (j+upper >=0) and (j+upper<ny)) : 
                            image[(i+left),(j+upper)] += kernel[i,j]*qt


            start_ind = end_ind

    return image
  
N = 20
x = y = z = hs= qts = mass = rhos = np.random.rand(N)
nx=ny=100
args = (x, y, qts,hs, nx, ny, 0.0, 1.0, 0.0, 1.0,1)

template_kernel_cpu(*args)
from compare_perf import compare_perf        

compare_perf(template_kernel_cpu, args)



from core_types import StructT, IncompatibleTypes, NoneT, Type, NoneType   
from ptr_type import ptr_type
from scalar_types import Int64, ScalarT, IntT, BoolT
from tuple_type import TupleT, repeat_tuple
from slice_type import SliceT
  
class ArrayT(StructT):
  
  def __init__(self, elt_type, rank):
    assert isinstance(elt_type, ScalarT), \
      "Can't create array with element type %s, currently only scalar elements supported" % \
      (elt_type,)
    self.elt_type = elt_type
    self.rank = rank 
    
    tuple_t = repeat_tuple(Int64, self.rank)

    self.shape_t = tuple_t
    self.strides_t = tuple_t
    self.ptr_t = ptr_type(self.elt_type)
    self._fields_ = [
      ('data', self.ptr_t),
      ('shape', tuple_t),
      ('strides', tuple_t),
      ('offset', Int64),
      ('size', Int64),
      # ('dtype', TypeValueT(self.elt_type))
    ]
    self._hash = hash( (elt_type, rank) )
  
  def children(self):
    yield self.elt_type
    yield self.shape_t
    yield self.strides_t
    yield self.ptr_t 
     
  def dtype(self):
    return self.elt_type.dtype

  def __str__(self):
    return "array%d(%s)" % (self.rank, self.elt_type)

  def __eq__(self, other):
    if self is other:
      return True 
    return other.__class__ is ArrayT and \
        self.elt_type == other.elt_type and \
        self.rank == other.rank

  def __hash__(self):
    return self._hash 

  def combine(self, other):
    if self == other:
      return self
    elif isinstance(other, ScalarT):
      elt_t = self.elt_type
      combined_elt_t = elt_t.combine(other)
      return make_array_type(combined_elt_t, self.rank)
    elif other.__class__ is ArrayT:
      assert self.rank == other.rank
      combined_elt_t = self.elt_type.combine(other.elt_type)
      return make_array_type(combined_elt_t, self.rank)
    else:
      raise IncompatibleTypes(self, other)

  def index_type(self, idx):
    """
    Given the type of my indices, what type of array will result?
    """

    if not isinstance(idx, Type):
      idx = idx.type

    if idx.__class__ is TupleT:
      indices = idx.elt_types
    else:
      indices = (idx,)

    n_indices = len(indices)
    n_required = self.rank
    if n_required > n_indices:
      n_missing = n_required - n_indices
      extra_indices = (NoneType,) * n_missing
      indices = indices + extra_indices

    # we lose one result dimension for each int in the index set
    result_rank = n_required
    for t in indices:
      if isinstance(t, (BoolT, IntT)):
        result_rank -= 1
      else:
        assert isinstance(t,  (TupleT, NoneT, SliceT, ArrayT, )),  "Unexpected index type: %s " % t
    if result_rank > 0:
      return make_array_type(self.elt_type, result_rank)
    else:
      return self.elt_type

_array_types = {}
def make_array_type(elt_t, rank):
  key = (elt_t, rank)
  arr_t = _array_types.get(key)
  if arr_t is None: 
    if rank == 0:  
      arr_t = elt_t 
    else: 
      arr_t = ArrayT(elt_t, rank)
    _array_types[key] = arr_t
  return arr_t

def elt_type(t):
  if t.__class__ is ArrayT:
    return t.elt_type
  else:
    return t

def elt_types(ts):
  return [elt_type(t) for t in ts]

def lower_rank(t, r):
  if t.__class__ is ArrayT:
    assert t.rank >= r
    return make_array_type(t.elt_type, t.rank - r)
  else:
    return t

def increase_rank(t, r):
  if t.__class__ is ArrayT:
    return make_array_type(t.elt_type, t.rank + r)
  else:
    return make_array_type(t, r)

def lower_ranks(arg_types, r):
  return [lower_rank(t, r) for t in arg_types]

def get_rank(t):
  if t.__class__ is ArrayT:
    return t.rank
  else:
    return 0

def rank(t):
  return get_rank(t)



import numpy as np 
from parakeet.testing_helpers import run_local_tests, expect_eq

import parakeet

def mk_sum(elt_t):
  array_t = parakeet.ndtypes.make_array_type(elt_t, 1)
  f, b, (x,) = parakeet.build_fn([array_t], elt_t)
  n = b.len(x)
  total, total_after, merge = b.loop_var('total', b.zero(elt_t))
  def loop_body(idx):
    b.assign(total_after, b.add(total, b.index(x,idx)))
  b.loop(0, n, loop_body, merge = merge)
  b.return_(total)
  return f 

def sum_i64(x):
  fn = mk_sum(parakeet.Int64)
  print fn 
  return parakeet.run_typed_fn(fn, (x,))
  
def sum_f64(x):
  fn = mk_sum(parakeet.Float64)
  return parakeet.run_typed_fn(fn, (x,))

def test_sum():
  expect_eq(sum_i64(np.array([1,2,3])), 6)
  expect_eq(sum_f64(np.array([-1.0, 1.0, 2.0])), 2.0)
  

if __name__ == '__main__':
 
  run_local_tests()  



import numpy as np
import parakeet 

from parakeet.testing_helpers import expect, run_local_tests 


int_vec = np.array([1,2,3])
float_vec = np.array([1.0,2.0,3.0])
bool_vec = np.array([True,False])

int_mat = np.array([int_vec, int_vec])
float_mat = np.array([float_vec, float_vec])
bool_mat = np.array([bool_vec, bool_vec])


def test_ones_like():
  expect(np.ones_like, [int_vec], np.ones_like(int_vec))

def test_ones_like_intmat():
  expect(np.ones_like, [int_mat], np.ones_like(int_mat))

def test_ones_like_floatvec():
  expect(np.ones_like, [float_vec], np.ones_like(float_vec))

def test_ones_like_floatmat():
  expect(np.ones_like, [float_mat], np.ones_like(float_mat))

def test_ones_like_boolvec():
  expect(np.ones_like, [bool_vec], np.ones_like(bool_vec))

def test_ones_like_boolmat():
  expect(np.ones_like, [bool_mat], np.ones_like(bool_mat))

def test_ones_like_boolvec_to_float():
  expect(np.ones_like, [bool_vec, np.dtype('float')], np.ones_like(bool_vec, dtype = np.dtype('float')))

def test_ones_like_boolmat_to_float():
  expect(np.ones_like, [bool_mat, np.dtype('float')], np.ones_like(bool_mat, dtype = np.dtype('float')))


if __name__ == '__main__':
  run_local_tests()


from parakeet.testing_helpers import expect, run_local_tests 

def sum_primes(n):
  total = 0
  count = 0
  for i in xrange(2, n):
    found_divisor = False
    for j in xrange(2,i-1):
      if i % j == 0:
        found_divisor = True
    if not found_divisor:
      total += i
      count += 1
  return total, count 

def test_sum_primes():
  expect(sum_primes, [20], sum_primes(20))  

if __name__ == '__main__':
  run_local_tests()




from .. import config, type_inference 
from ..analysis import contains_loops 
from ..ndtypes import type_conv, Type, typeof  
from ..syntax import UntypedFn, TypedFn, ActualArgs
from ..transforms import pipeline

from .. import c_backend
from .. import openmp_backend 

import ast_conversion

# get types of all inputs
def _typeof(arg):
  try:
    return typeof(arg)
  except:
    if hasattr(arg, 'type') and isinstance(arg.type, Type): 
      return arg.type
    else:
      raise 
  
def prepare_args(fn, args, kwargs):
  """
  Fetch the function's nonlocals and return an ActualArgs object of both the arg
  values and their types
  """
  #assert not isinstance(fn, TypedFn), "[prepare_args] Only works for untyped functions"
  if not isinstance(fn, UntypedFn):
    fn = ast_conversion.translate_function_value(fn)   
  nonlocals = tuple(fn.python_nonlocals())
  arg_values = ActualArgs(nonlocals + tuple(args), kwargs)
  arg_types = arg_values.transform(_typeof)
  return arg_values, arg_types
  

def specialize(untyped, args, kwargs = {}, optimize = False):
  """
  Translate, specialize and begin to optimize the given function for the types
  of the supplies arguments.

  Return the untyped and typed representations, along with all the
  arguments in a linear order. 
  """

  if not isinstance(untyped, UntypedFn):
    untyped = ast_conversion.translate_function_value(untyped)
       
  arg_values, arg_types = prepare_args(untyped, args, kwargs)
  
  # convert the awkward mix of positional, named, and starargs 
  # into a positional sequence of arguments
  linear_args = untyped.args.linearize_without_defaults(arg_values)
  
  # propagate types through function representation and all
  # other functions it calls
   
  typed_fn = type_inference.specialize(untyped, arg_types)
  if optimize: 
    from .. transforms.pipeline import normalize 
    # apply high level optimizations 
    typed_fn = normalize.apply(typed_fn)
  return typed_fn, linear_args 

def run_typed_fn(fn, args, backend = None):
  actual_types = tuple(type_conv.typeof(arg) for arg in  args)
  expected_types = fn.input_types
  assert actual_types == expected_types, \
    "Arg type mismatch, expected %s but got %s" % \
    (expected_types, actual_types)
  
  if backend is None:
    backend = config.backend
    
  if backend == 'c':
    return c_backend.run(fn, args)
   
  elif backend == 'openmp':
    return openmp_backend.run(fn, args)
  
  elif backend == 'cuda':
    # only selectively import cuda_backend since it required PyCUDA
    from .. import cuda_backend 
    return cuda_backend.run(fn, args)
  
  elif backend == 'llvm':
    from ..llvm_backend.llvm_context import global_context
    from ..llvm_backend import generic_value_to_python 
    from ..llvm_backend import ctypes_to_generic_value, compile_fn 
    lowered_fn = pipeline.lowering.apply(fn)
    llvm_fn = compile_fn(lowered_fn).llvm_fn

    ctypes_inputs = [t.from_python(v) 
                   for (v,t) 
                   in zip(args, expected_types)]
    gv_inputs = [ctypes_to_generic_value(cv, t) 
                 for (cv,t) 
                 in zip(ctypes_inputs, expected_types)]
    exec_engine = global_context.exec_engine
    gv_return = exec_engine.run_function(llvm_fn, gv_inputs)
    return generic_value_to_python(gv_return, fn.return_type)

  elif backend == "interp":
    from .. import interp 
    fn = pipeline.loopify(fn)
    return interp.eval_fn(fn, args)
  
  else:
    assert False, "Unknown backend %s" % backend 

def run_untyped_fn(fn, args, kwargs = None, backend = None):
  assert isinstance(fn, UntypedFn)
  if kwargs is None:
    kwargs = {}
  typed_fn, linear_args = specialize(fn, args, kwargs)
  return run_typed_fn(typed_fn, linear_args, backend)

def run_python_ast(fn_name, fn_args, fn_body, globals_dict, 
                     arg_values, kwarg_values = None, backend = None):
  """
  Instead of giving Parakeet a function to parse, you can construct a Python 
  AST yourself and then have that converted into a Typed Parakeet function
  """
  untyped = ast_conversion.translate_function_ast(fn_name, fn_args, fn_body, globals_dict)
  return run_untyped_fn(untyped, arg_values, kwarg_values, backend)


def run_python_fn(fn, args, kwargs = None, backend = None):
  """
  Given a python function, run it in Parakeet on the supplied args
  """
  # translate from the Python AST to Parakeet's untyped format
  untyped = ast_conversion.translate_function_value(fn)
  return run_untyped_fn(untyped, args, kwargs, backend)
  



import type_conv

from core_types import StructT, ImmutableT, IncompatibleTypes

class SliceT(StructT, ImmutableT):
  def __init__(self, start_type, stop_type, step_type):
    self.start_type = start_type
    self.stop_type = stop_type
    self.step_type = step_type
    self._fields_ = [('start',start_type), ('stop', stop_type), ('step', step_type)]
    self._hash = hash((self.start_type, self.stop_type, self.step_type))
  
  def children(self):
    yield self.start_type
    yield self.stop_type
    yield self.step_type
    
  def __eq__(self, other):
    return self is other or \
      (other.__class__ is SliceT and
       self.start_type == other.start_type and
       self.stop_type == other.stop_type and
       self.step_type == other.step_type)

  def __hash__(self):
    return self._hash

  def combine(self, other):
    if self == other: return self
    else:raise IncompatibleTypes(self, other)

  def __str__(self):
    return "SliceT(%s, %s, %s)" % (self.start_type,
                                   self.stop_type,
                                   self.step_type)

  def __repr__(self):
    return str(self)

_slice_type_cache = {}
def make_slice_type(start_t, stop_t, step_t):
  key = (start_t, stop_t, step_t)
  if key in _slice_type_cache:
    return _slice_type_cache[key]
  else:
    t = SliceT(start_t, stop_t, step_t)
    _slice_type_cache[key] = t
    return t

def typeof_slice(s):
  start_type = type_conv.typeof(s.start)
  stop_type = type_conv.typeof(s.stop)
  step_type = type_conv.typeof(s.step)
  return make_slice_type(start_type, stop_type, step_type)

type_conv.register(slice, SliceT, typeof_slice)



from dsltools import Node 

from .. ndtypes import NoneT

class Expr(object):
  # _members = ['type', 'source_info']
  
  def __str__(self):
    fields = ["%s = %s" % (k,v) for (k,v) in self.__dict__.iteritems()]
    return "%s(%s)" % (self.__class__.__name__, ", ".join(fields))
  
  def __repr__(self):
    return str(self)
  
  @classmethod
  def node_type(cls):
    return cls.__name__
  
  def children(self):
    for v in self.itervalues():
      if v and isinstance(v, Expr):
        yield v
      elif isinstance(v, (list,tuple)):
        for child in v:
          if isinstance(child, Expr):
            yield child

  def short_str(self):
    return str(self)

  def __eq__(self, other):
    return (self.__class__ is other.__class__) and \
      self.type == other.type and \
      all(c1 == c2 for (c1,c2) in zip(self.children(), other.children()))
    
  def __ne__(self, other):
    return not (self == other)
  
  def hash(self):
    elts = tuple(self.children())
    return hash(elts)
   
class Const(Expr):
  def __init__(self, value, type = None, source_info = None):
    self.value = value 
    self.type = type 
    self.source_info = source_info 
  
  def children(self):
    return ()

  def short_str(self):
    return str(self.value)

  def __repr__(self):
    if self.type and not isinstance(self.type, NoneT):
      return "%s : %s" % (self.value, self.type)
    else:
      return str(self.value)

  def __str__(self):
    return repr(self)

  def __hash__(self):
    return hash(self.value)

  def __eq__(self, other):
    return other.__class__ is Const and \
           self.value == other.value and \
           self.type == other.type
    
  def __ne__(self, other):
    return other.__class__ is not Const or \
           self.value != other.value or \
           self.type != other.type

class Var(Expr):
  def __init__(self, name, type = None, source_info = None):
    assert name is not None 
    self.name = name
    self.type = type 
    self.source_info = source_info 

  # it's possible to speed things up a little
  # by by-passing the Node construct 
  #_members = ['name']
  #def node_init(self):
  #  assert self.name is not None
    
  def short_str(self):
    return self.name

  def __repr__(self):
    if hasattr(self, 'type'):
      return "%s : %s" % (self.name, self.type)
    else:
      return "%s" % self.name

  def __str__(self):
    return self.name

  def __hash__(self):
    return hash(self.name)

  def __eq__(self, other):
    return other.__class__  is Var and \
           self.name == other.name and \
           self.type == other.type


  def __ne__(self, other):
    if other.__class__ is not Var:
      return True 
    if self.name != other.name:
      return True
    return self.type != other.type 
  
  def children(self):
    return ()

class Attribute(Expr):
  def __init__(self, value, name, type = None, source_info = None):
    self.value = value 
    self.name = name 
    self.type = type 
    self.source_info = source_info 

  def children(self):
    yield self.value

  def __str__(self):
    return "attr(%s, '%s')" % (self.value, self.name)

  def __hash__(self):
    return hash ((self.value, self.name))

  def __eq__(self, other):
    return other.__class__ is Attribute and \
           self.name == other.name and \
           self.value == other.value


class Closure(Expr):
  """Create a closure which points to a global fn with a list of partial args"""
  def __init__(self, fn, args, type = None, source_info = None):
    self.fn = fn 
    self.args = args 
    self.type = type 
    self.source_info = source_info 

  def __str__(self):
    fn_str = str(self.fn) #self.fn.name if hasattr(self.fn, 'name') else str(self.fn)
    args_str = ",".join(str(arg) + ":" + str(arg.type) for arg in self.args)
    return "Closure(%s, fixed_args = {%s})" % (fn_str, args_str)

  def node_init(self):
    self.args = tuple(self.args)
    #self.return_type = self.fn.return_type
    #self.input_types = self.fn.input_types[len(self.args):]

  def children(self):
    if isinstance(self.fn, Expr):
      yield self.fn
    for arg in self.args:
      yield arg

  def __hash__(self):
    return hash((self.fn, tuple(self.args)))

class Call(Expr):
  def __init__(self, fn, args, type = None, source_info = None):
    self.fn = fn 
    self.args = args 
    self.type = type 
    self.source_info = source_info 
    
  def __str__(self):
    #if isinstance(self.fn, (Fn, TypedFn)):
    #  fn_name = self.fn.name
    #else:
    fn_name = str(self.fn)
    if self.args.__class__.__name__ == 'ActualArgs':
      arg_str = str(self.args)
    else:
      arg_str = ", ".join(str(arg) for arg in self.args)
    return "%s(%s)" % (fn_name, arg_str)

  def __repr__(self):
    return str(self)

  def children(self):
    yield self.fn
    for arg in self.args:
      yield arg

  def __hash__(self):
    return hash((self.fn, tuple(self.args)))


class PrimCall(Expr):
  """
  Call a primitive function, the "prim" field should be a prims.Prim object
  """
  def __init__(self, prim, args, type = None, source_info = None):
    self.prim = prim 
    self.args = args 
    self.type = type 
    self.source_info = source_info 
    
  #_members = ['prim', 'args']
  
  def _arg_str(self, i):
    arg = self.args[i]
    if arg.__class__ is PrimCall:
      return "(%s)" % arg
    else:
      return str(arg)
  
  def __hash__(self):
    self.args = tuple(self.args)
    return hash((self.prim, self.args))
    
  def __eq__(self, other):
    if other.__class__ is not PrimCall:
      return False 
    
    if other.prim != self.prim:
      return False
    
    my_args = self.args
    other_args = other.args
    n = len(my_args)
    if n != len(other_args):
      return False
    
    for (i,arg) in enumerate(my_args):
      other_arg = other_args[i]
      if arg != other_arg:
        return False
    return True
  
  def __ne__(self, other):
    return not self==other
  
  def __repr__(self):
    if self.prim.symbol:
      if len(self.args) == 1:
        return "%s%s" % (self.prim.symbol, self._arg_str(0))
      else:
        assert len(self.args) == 2
        return "%s %s %s" % (self._arg_str(0), self.prim.symbol, self._arg_str(1))
    else:
      arg_strings = [self._arg_str(i) for i in xrange(len(self.args))]
      combined = ", ".join(arg_strings)
      return "prim<%s>(%s)" % (self.prim.name, combined)

  def __str__(self):
    return repr(self)

  def node_init(self):
    self.args = tuple(self.args)

  def children(self):
    return self.args
  


class ClosureElt(Expr):
  def __init__(self, closure, index, type = None, source_info = None):
    self.closure = closure 
    self.index = index 
    self.type = type 
    self.source_info = source_info 
    
  def __str__(self):
    return "ClosureElt(%s, %d)" % (self.closure, self.index)

  def children(self):
    return (self.closure,)

  def __hash__(self):
    return hash((self.closure, self.index))

class Cast(Expr):
  def __init__(self, value, type, source_info = None):
    self.value = value 
    self.type = type 
    self.source_info = source_info 

  def children(self):
    yield self.value 

  def __hash__(self):
    return hash(self.value)
  
  def __str__(self):
    return "Cast(%s : %s)" % (self.value, self.type) 

class Select(Expr):
  def __init__(self, cond, true_value, false_value, type = None, source_info = None):
    self.cond = cond 
    self.true_value = true_value 
    self.false_value = false_value 
    self.type = type 
    self.source_info = source_info 
    
  # _members = ['cond', 'true_value', 'false_value']
  
  def __hash__(self):
    return hash((self.cond, self.true_value, self.false_value))
  
  def __str__(self):
    return "Select(%s, %s, %s)" % (self.cond, self.true_value, self.false_value)
  
  def children(self):
    yield self.cond 
    yield self.true_value 
    yield self.false_value 
  
  




from dsltools import ScopedDict

from .. import config 
from ..analysis.verify import verify
from ..ndtypes import (NoneT, SliceT, ScalarT, ArrayT, ClosureT, TupleT, PtrT, 
                       StructT, make_closure_type) 
from ..syntax import TypedFn
from transform import Transform


class RecursiveApply(Transform):  
  cache = ScopedDict()
  
  def __init__(self, transform_to_apply = None):
    Transform.__init__(self)
    self.transform = transform_to_apply
    self.cache.push()
  
  def __del__(self):
    self.cache.pop()
  
  
  first_order_types = (SliceT, NoneT, ArrayT, PtrT)
  
  def contains_function_type(self, t):
    c = t.__class__ 
    if c is ClosureT: 
      return True 
    elif c is TupleT: 
      return any(self.contains_function_type(elt_t) for elt_t in t.elt_types)
    elif c in self.first_order_types or isinstance(t, ScalarT):
      return False 
    else:
      assert isinstance(t, StructT), "Unexpected type %s" % t
      return any(self.contains_function_type(field_t) for field_t in t.field_types)
    
  def transform_type(self, t):
    c = t.__class__ 
    if c is ClosureT:
      old_fn = t.fn 
      if isinstance(old_fn, TypedFn):
        new_fn = self.transform_TypedFn(old_fn)
        if new_fn is not old_fn:
          return make_closure_type(new_fn, t.arg_types)
    elif c is TupleT and self.contains_function_type(t):
      new_elt_types = []
      for elt_t in t.elt_types:
        new_elt_types.append(self.transform_type(elt_t))
      return TupleT(tuple(new_elt_types))
    # if it's neither a closure nor a structure which could contain closures, 
    # just return it 
    return t 
    
      
  def transform_TypedFn(self, expr):
    key = expr.cache_key
    if key in self.cache:
      return self.cache[key]  
    new_fn = self.transform.apply(expr)
    if config.opt_verify:
      try: 
        verify(new_fn)
      except:
        print "[RecursiveApply] Error after applying %s to function %s" % (self.transform, expr)
        raise 
    self.cache[key] = new_fn 
    return new_fn 
  
  def transform_Closure(self, expr):
    args = self.transform_expr_list(expr.args)
    new_fn = self.transform_expr(expr.fn)
    return self.closure(new_fn, args)
  
  def transform_Var(self, expr):
    expr.type = self.transform_type(expr.type)
    return expr
  
  def transform_Assign(self, stmt):
    """
    If we have an assignment like 
      a : (Fn1T, Fn2T) = (fn1, fn2)
    we might need to change it to 
      a : (Fn1T', Fn2T') = (fn1', fn2') if the RHS functions get updated
    """ 
    stmt.rhs = self.transform_expr(stmt.rhs)
    stmt.lhs.type = self.transform_type(stmt.lhs.type)
    return stmt 
  
  def pre_apply(self, fn):
    if self.transform is None:
      self.transform = fn.created_by 
    
    assert self.transform is not None, "No transform specified for RecursiveApply"
    fn.input_types = tuple(self.transform_type(t) for t in fn.input_types)
    fn.return_type = self.transform_type(fn.return_type)
    for k,t in fn.type_env.items():
      fn.type_env[k] = self.transform_type(t)
    return fn   




import config 
def get_cuda_devices(compute_capability = config.compute_capability):
  try:
    import pycuda.autoinit
    import pycuda.driver  
  except:
    return []
  devices = [pycuda.driver.Device(i) for i in xrange(pycuda.driver.Device.count())]
  return [d for d in devices if d.compute_capability() >= compute_capability]

def device_id(cuda_device):
  import pycuda.driver 
  return cuda_device.get_attribute(pycuda.driver.device_attribute.PCI_DEVICE_ID)

def display_attached(cuda_device):
  import pycuda.driver 
  return cuda_device.get_attribute(pycuda.driver.device_attribute.KERNEL_EXEC_TIMEOUT) == 1  
  
def num_multiprocessors(cuda_device):
  import pycuda.driver
  return cuda_device.get_attribute(pycuda.driver.device_attribute.MULTIPROCESSOR_COUNT)
  

def best_cuda_device(compute_capability = config.compute_capability):
  devices = get_cuda_devices(compute_capability)
  if len(devices) == 0:
    return None
  
  best_device = devices[0]
  for d in devices[1:]:
    if display_attached(best_device) and not display_attached(d):
      best_device = d 
    elif num_multiprocessors(d) > num_multiprocessors(best_device):
      best_device = d
    elif d.total_memory() > best_device.total_memory():
      best_device = d
  return best_device    

def has_gpu():
  return best_cuda_device() is not None

    
  
  
  


import numpy as np 
from parakeet import jit 
from parakeet.testing_helpers import expect, run_local_tests

@jit 
def range1(a):
  return range(a)

def test_range1():
  expect(range1, [10], np.arange(10))

@jit 
def range2(a,b):
  return range(a,b)

def test_range2():
  expect(range2, [10, 20], np.arange(10,20))

@jit
def range3(a,b,c):
  return range(a,b,c)
  
def test_range3():
  expect(range3, [20,45,3], np.arange(20,45,3))

if __name__ == '__main__':
  run_local_tests()



import numpy as np

from ..frontend import macro, jit 
from ..syntax import (UntypedFn, Return, Cast, TypedFn, TypeValue, Array, Tuple, Expr, Closure, 
                      TupleProj) 
from ..syntax.helpers import get_types, make_tuple  
from ..ndtypes import (type_conv, TypeValueT, ArrayT, IntT, 
                       make_tuple_type, TupleT, ScalarT, ClosureT, FnT, Type)


def _get_type(dtype):
  """
  Defensively try to extract a scalar type from any wacky thing
  that might get passed in as a 'dtype' to an array constructor
  """
  if isinstance(dtype, macro):
    dtype = dtype.as_fn()
  
  while isinstance(dtype, jit):
    dtype = dtype.f 
  
  while isinstance(dtype, Expr):
    if isinstance(dtype, UntypedFn):
      if len(dtype.body) == 1:
        stmt = dtype.body[0]
        if stmt.__class__ is Return:
          expr = stmt.value 
          if expr.__class__ is Cast:
            dtype = expr.type 
            break
      assert False, "Don't know how to convert function %s into Parakeet type" % dtype
    elif isinstance(dtype, TypedFn):
      dtype = dtype.return_type
    if isinstance(dtype, TypeValue):
      dtype = dtype.type_value  
    elif isinstance(dtype.type, TypeValueT):
      dtype = dtype.type
    elif isinstance(dtype, Closure):
      dtype = dtype.fn
    elif isinstance(dtype.type, (ClosureT, FnT)):
      dtype = dtype.type.fn
    else:
      assert False, "Don't know how to turn %s : %s into Parakeet type" % (dtype, dtype.type)  


  if isinstance(dtype, Type):
    if isinstance(dtype, TypeValueT):
      return dtype.type
    else:
      return dtype 
    
  elif isinstance(dtype, (np.dtype, type)):

    return type_conv.equiv_type(dtype)
  elif isinstance(dtype, str):
    return type_conv.equiv_type(np.dtype(dtype))
  assert False, "Don't know how to turn %s into Parakeet type" % dtype  

def _get_shape(value):
  """
  User might pass a scalar, a tuple of integer values, or a literal array of integers
  as the 'shape' argument of array constructor. Normalize those all into a tuple
  """
  if isinstance(value.type, ArrayT):
    assert value.__class__ is Array, "Don't know how to convert %s into tuple" % value
    elts = value.elts 
    
    elt_types = get_types(elts)
    assert all(isinstance(t, IntT) for t in elt_types), \
      "Shape elements must be integers, not %s" % elt_types
    return Tuple(elts = elts, type = make_tuple_type(elt_types))
  elif isinstance(value.type, TupleT):
    assert all(isinstance(t, ScalarT) for t in value.type.elt_types), \
      "Shape tuple %s : %s has non-scalar elements" % (value, value.type)
    return value
  elif isinstance(value.type, ScalarT):
    assert isinstance(value.type, IntT), \
      "Can't make shape tuple from non-integer scalar %s : %s" % (value, value.type)
    return make_tuple((value,))
  assert False, "Can't make shape tuple from value %s : %s" % (value, value.type)
  
def _get_tuple_elts(expr, cast_type = None):
  elts = []
  for i, tuple_elt_t in enumerate(expr.type.elt_types):
    if expr.__class__ is Tuple:
      tuple_elt = expr.elts[i]
    else:
      tuple_elt = TupleProj(expr, i, type = tuple_elt_t)
    if cast_type is not None and tuple_elt_t != cast_type:
      tuple_elt = Cast(tuple_elt, type = cast_type)
    elts.append(tuple_elt)
  return tuple(elts)  
    


from parakeet.testing_helpers import run_local_tests, expect_eq

def square(x):
  return np.square(x)

def test_square_scalars():
  expect_eq(square(1), 1)
  expect_eq(square(2), 4)
  expect_eq(square(-2), 4)
  expect_eq(square(True), True)
  expect_eq(square(False), False)
  expect_eq(square(-4.0), 16.0)

import numpy as np 
intvec = np.array([-3,-2,-1,0,1,2,3])
floatvec = intvec.astype('float')

def test_square_vectors():
  expect_eq(square(intvec), intvec*intvec)
  expect_eq(square(floatvec), floatvec*floatvec)

intmat = np.array([intvec, intvec])
floatmat = intmat.astype('float')

def test_square_matrices():
  expect_eq(square(intmat), intmat*intmat)
  expect_eq(square(floatmat), floatmat*floatmat)


if __name__ == '__main__':
  run_local_tests()



from .. import prims 

from .. frontend import jit,  typed_macro 
from .. ndtypes import make_tuple_type, TupleT, ArrayT
from ..syntax import Tuple, Array 

@jit 
def builtin_or(x, y):
  return x or y

@jit 
def builtin_and(x, y):
  return x and y


@typed_macro 
def builtin_tuple(xt):
  if isinstance(xt.type, TupleT):
    return xt 
  else:
    assert isinstance(xt.type, ArrayT), "Can't create type from %s" % (xt.type,)
    assert isinstance(xt, Array), "Can only create tuple from array of const length"
    elt_types = [e.type for e in xt.elts]
    tuple_t = make_tuple_type(elt_types)
    return Tuple(xt.elts, type = tuple_t)
  




from .. import names 
  
from .. syntax import (Expr, Var, Const, Return, UntypedFn, FormalArgs, DelayUntilTyped,  
                       const, is_python_constant)

from run_function import run_untyped_fn, run_typed_fn, specialize 

class jit(object):
  def __init__(self, f):
    self.f = f
    self.fn = f
    self.untyped = None 

  def __call__(self, *args, **kwargs):
    if '_backend' in kwargs:
      backend_name = kwargs['_backend']
      del kwargs['_backend']
    else:
      backend_name = None
    
    if self.untyped is None:
      import ast_conversion 
      self.untyped = ast_conversion.translate_function_value(self.fn)
    
    typed_fn, linear_args = specialize(self.untyped, args, kwargs)
    return run_typed_fn(typed_fn, linear_args, backend_name)


class macro(object):
  def __init__(self, f, static_names = set([]), call_from_python = None):
    self.f = f
    self.fn = f 
    self.argcount = self.f.func_code.co_argcount 
    self.varnames = self.f.func_code.co_varnames 
    self.n_args = self.f.func_code.co_argcount
    self.defaults = self.f.func_defaults
    
    n_defaults = len(self.defaults) if self.defaults else 0 
    self.defaults_dictionary = {}
    n_pos = self.n_args - n_defaults
    for i, default_name in enumerate(self.varnames[n_pos:n_pos+n_defaults]):
      self.defaults_dictionary[default_name] = self.defaults[i]
    
    self.static_names = static_names
    self.wrappers = {}
    self.call_from_python = call_from_python
    if hasattr(self.f, "__name__"):
      self.name = f.__name__
    elif hasattr(self.f, "name"):
      self.name = self.f.name 
    else:
      self.name = "macro"

  
  _macro_wrapper_cache = {}
  def _create_wrapper(self, n_pos, static_pairs, dynamic_keywords):
    static_pairs = tuple(static_pairs)
    
    key = (n_pos, static_pairs, dynamic_keywords)
    if key in self.wrappers:
      return self.wrappers[key]
    
    args = FormalArgs()
    pos_vars = []
    keyword_vars = {}
    
    for i in xrange(n_pos):
      if i <  self.argcount: 
        raw_name = self.varnames[i]
      else:
        raw_name = "input_%d" % i
      local_name = names.fresh(raw_name)
      args.add_positional(local_name)
      pos_vars.append(Var(local_name))

      
    import ast_conversion
    for visible_name in dynamic_keywords:
      local_name = names.fresh(visible_name)
      args.add_positional(local_name, visible_name)
      keyword_vars[visible_name] = Var(local_name)
      if visible_name in self.defaults_dictionary:
        default_value = self.defaults_dictionary[visible_name]
        parakeet_value = ast_conversion.value_to_syntax(default_value)
        args.defaults[local_name] = parakeet_value

    for (static_name, value) in static_pairs:
      if isinstance(value, Expr):
        assert isinstance(value, Const)
        keyword_vars[static_name] = value
      elif value is not None:
        assert is_python_constant(value), \
            "Unexpected type for static/staged value: %s : %s" % \
            (value, type(value))
        keyword_vars[static_name] = const(value)


    result_expr = self.f(*pos_vars, **keyword_vars)
    body = [Return(result_expr)]
    wrapper_name = "%s_wrapper_%d_%d" % (self.name, n_pos,
                                         len(dynamic_keywords))
    wrapper_name = names.fresh(wrapper_name)
    untyped = UntypedFn(name = wrapper_name, args = args, body = body)
    self.wrappers[key] = untyped
    return untyped 
  
  def as_fn(self):
    n_default = len(self.defaults) if self.defaults else 0 
    n_pos = self.n_args - n_default 
    arg_varnames = self.varnames[:self.argcount]
    keyword_names = arg_varnames[n_pos:]
    wrapper = self._create_wrapper(n_pos, [], keyword_names)
    return wrapper 
    
  def __call__(self, *args, **kwargs):

    if self.call_from_python is not None:
      return self.call_from_python(*args, **kwargs)
    
    if '_backend' in kwargs:
      backend_name = kwargs['_backend']
      del kwargs['_backend']
    else:
      backend_name = None

    n_pos = len(args)
    keywords = kwargs.keys()
    static_pairs = tuple((k,kwargs.get(k)) for k in self.static_names)
   
    dynamic_keywords = tuple(k for k in keywords
                               if k not in self.static_names)

     
    untyped = self._create_wrapper(n_pos, static_pairs, dynamic_keywords)

    dynamic_kwargs = dict( (k, kwargs[k]) for k in dynamic_keywords)
    return run_untyped_fn(untyped, args, dynamic_kwargs, backend = backend_name)
    

  def transform(self, args, kwargs = {}):
    for arg in args:
      assert isinstance(arg, Expr), \
          "Macros can only take syntax nodes as arguments, got %s" % (arg,)
    for (name,arg) in kwargs.iteritems():
      assert isinstance(arg, Expr), \
          "Macros can only take syntax nodes as arguments, got %s = %s" % \
          (name, arg)
    result = self.f(*args, **kwargs)
    assert isinstance(result, Expr), \
        "Expected macro %s to return syntax expression, got %s" % \
        (self.f, result)
    return result

  def __str__(self):
    return "macro(%s)" % self.name

class typed_macro(macro):
  def __init__(self, f, *args, **kwargs):
    macro.__init__(self, f, *args, **kwargs)
    def delayed(*macro_args, **macro_kwargs):
      return DelayUntilTyped(values = macro_args, keywords = macro_kwargs, fn = f)
    self.f = delayed 
    

class staged_macro(object):
  def __init__(self, *static_names, **kwargs):
    self.static_names = tuple(static_names)

    self.call_from_python = kwargs.get('call_from_python')
    assert kwargs.keys() in [[], ['call_from_python']], \
        "Unknown keywords: %s" % kwargs.keys()

  def __call__(self, fn):
    return macro(fn, 
                 self.static_names,
                 call_from_python = self.call_from_python)

axis_macro = staged_macro("axis")


import numpy as np 
import parakeet 
import parakeet.c_backend

from parakeet.testing_helpers import run_local_tests, expect 
def wrap(pos, offset, bound):
    return ( pos + offset ) % bound

def clamp(pos, offset, bound):
    return min(bound-1,max(0,pos+offset))

def reflect(pos, offset, bound):
    idx = pos+offset
    return min(2*(bound-1)-idx,max(idx,-idx))


def python_local_maxima(data, wsize, mode=wrap):
  result = np.ones(shape=data.shape,dtype='bool')
  for pos in np.ndindex(data.shape):
    myval = data[pos]  
    for offset in np.ndindex(wsize):
      neighbor_idx = tuple(mode(p, o-w/2, w) for (p, o, w) in zip(pos, offset, wsize))
      result[pos] &= (data[neighbor_idx] <= myval)
  return result 

@parakeet.jit 
def local_maxima(data, wsize, mode=wrap):
  def is_max(pos):
    def is_smaller_neighbor(offset):
      neighbor_idx = tuple(mode(p, o-w/2, w) for (p, o, w) in zip(pos, offset, wsize))
      return data[neighbor_idx] <= data[pos]
    return np.all(parakeet.imap(is_smaller_neighbor, wsize))
  return parakeet.imap(is_max, data.shape)
  

shape = (4,3,3,3)
x = np.random.randn(*shape)
wsize = (3,3,3,3)

def test_local_maxima():
  expect(local_maxima, [x, wsize], python_local_maxima(x, wsize))
  


if __name__ == '__main__':
    run_local_tests()
  




import llvm.core as llc   

import llvm_types
from .. import prims
from .. ndtypes import Float32, Float64, Int32, Int64
from llvm_context import global_context

  

signed_int_comparisons = {
  prims.equal : llc.ICMP_EQ, 
  prims.not_equal : llc.ICMP_NE, 
  prims.greater : llc.ICMP_SGT, 
  prims.greater_equal : llc.ICMP_SGE, 
  prims.less : llc.ICMP_SLT, 
  prims.less_equal : llc.ICMP_SLE
}

unsigned_int_comparisons = {
  prims.equal : llc.ICMP_EQ, 
  prims.not_equal : llc.ICMP_NE, 
  prims.greater : llc.ICMP_UGT, 
  prims.greater_equal : llc.ICMP_UGE, 
  prims.less : llc.ICMP_ULT, 
  prims.less_equal : llc.ICMP_ULE
}  

float_comparisons = { 
  prims.equal : llc.FCMP_OEQ, 
  prims.not_equal : llc.FCMP_ONE, 
  prims.greater : llc.FCMP_OGT, 
  prims.greater_equal : llc.FCMP_OGE, 
  prims.less : llc.FCMP_OLT, 
  prims.less_equal : llc.FCMP_OLE                     
}


signed_binops = { 
  prims.add : 'add',  
  prims.subtract : 'sub', 
  prims.multiply : 'mul', 
  prims.divide : 'sdiv',
}

unsigned_binops = { 
  prims.add : 'add',  
  prims.subtract : 'sub', 
  prims.multiply : 'mul', 
  prims.divide : 'udiv', 
  prims.mod : 'urem', 
}

float_binops = { 
  prims.add : 'fadd',  
  prims.subtract : 'fsub', 
  prims.multiply : 'fmul', 
  prims.divide : 'fdiv', 
}

# Note: there is no division instruction between booleans
# so b1 / b2 should be translated to int(b1) / int(b2) 
bool_binops = { 
  prims.add : 'or_',
  prims.multiply : 'and_', 
  prims.subtract : 'xor',
  prims.divide : 'and_',
  
  prims.mod : 'urem'
  #prims.logical_not : 'not_'
}

int32_fn_t = llc.Type.function(llvm_types.int32_t, [llvm_types.int32_t])
int64_fn_t = llc.Type.function(llvm_types.int64_t, [llvm_types.int64_t])
float32_fn_t = llc.Type.function(llvm_types.float32_t, [llvm_types.float32_t])
float64_fn_t = llc.Type.function(llvm_types.float64_t, [llvm_types.float64_t])

binary_float32_fn_t = llc.Type.function(llvm_types.float32_t, 
                                        [llvm_types.float32_t, llvm_types.float32_t])
binary_float64_fn_t = llc.Type.function(llvm_types.float64_t, 
                                        [llvm_types.float64_t, llvm_types.float64_t])

def float32_fn(name):
  return global_context.module.add_function(float32_fn_t, name)

def float64_fn(name):
  return global_context.module.add_function(float64_fn_t, name)



#import llvmmath 
#mathlib = llvmmath.get_default_math_lib()
#linker = llvmmath.linking.get_linker(mathlib)

_llvm_intrinsics = set(['sqrt', 
                        'powi', 
                        'sin', 
                        'cos',
                        'pow',
                        'exp', 
                        'exp2', 
                        'log', 
                        'log10', 
                        'log2', 
                        'fma', 
                        'fabs', 
                        'floor', 
                        'ceil', 
                        'trunc', 
                        'rint', 
                        'nearbyint',                         
                      ])

from ..c_backend.c_prims import _float_fn_names

def get_float_op(prim, t, _float_decls = {}):
  key = (prim, t)
  if key in _float_decls:
    return _float_decls[key]
  
  assert prim in _float_fn_names, \
    "Unsupported float primitive %s" % prim 
  assert t in (Float32, Float64), \
    "Invalid type %s, expected Float32 or Float64" % t
  
  prim_name = _float_fn_names[prim]
  
  if t == Float32:
    fn_t = float32_fn_t if prim.nin == 1 else binary_float32_fn_t
    if prim_name in _llvm_intrinsics:
      fn_name = "llvm.%s.f32" % prim_name
    else:
      fn_name = prim_name + "f" 
  else:
    assert t == Float64
    fn_t = float64_fn_t if prim.nin == 1 else binary_float64_fn_t
    if prim_name in _llvm_intrinsics:
      fn_name = "llvm.%s.f64" % prim_name 
    else:
      fn_name = prim_name 
  llvm_value = global_context.module.get_or_insert_function(fn_t, fn_name)
  llvm_value.add_attribute(llc.ATTR_NO_UNWIND)
  llvm_value.add_attribute(llc.ATTR_READONLY)
  _float_decls[key] = llvm_value
  return llvm_value 




import ctypes

import type_conv

from core_types import IncompatibleTypes, StructT, ImmutableT


class TupleT(StructT, ImmutableT):
  rank = 0
  def __init__(self, elt_types):
    if elt_types is None:
      elt_types = ()
    else:
      elt_types = tuple(elt_types)
    self.elt_types = elt_types
    self._fields_ = [("elt%d" % i, t) for (i,t) in enumerate(self.elt_types)]
    self._hash = hash(elt_types)
    
  # signals to type inference algorithm to pass the value of an index into
  # index_type rather than its type
  static_indexing = True

  def children(self):
    return self.elt_types
  
  def from_python(self, python_tuple, _keep_forever = []):
    # _keep_forever.append(python_tuple)
    converted_elts = []
    for elt in python_tuple:
      parakeet_type = type_conv.typeof(elt)
      c_elt = parakeet_type.from_python(elt)

      if isinstance(parakeet_type, StructT):
        c_elt = ctypes.pointer(c_elt)
      converted_elts.append(c_elt)
    return self.ctypes_repr(*converted_elts)

  def to_python(self, struct_obj):
    elt_values = []
    for (field_name, field_type) in self._fields_:
      c_elt = getattr(struct_obj, field_name)
      if isinstance(field_type, StructT):
        c_elt = c_elt.contents

      py_elt = field_type.to_python(c_elt)

      elt_values.append(py_elt)
    return tuple(elt_values)

  def dtype(self):
    raise RuntimeError("Do tuples have dtypes?")

  def __eq__(self, other):
    if self is other: return True 
    return other.__class__ is TupleT and self.elt_types == other.elt_types

  def __hash__(self):
    return self._hash 

  def __str__(self):
    return "tuple(%s)" % ", ".join([str(t) for t in self.elt_types])

  def __repr__(self):
    return str(self)

  def __len__(self):
    return len(self.elt_types)

  def __iter__(self):
    return iter(self.elt_types)
  
  def __getitem__(self, idx):
    assert isinstance(idx, (int,long)), "Invalid index for TupleT: %s" % idx 
    return self.elt_types[idx]

  def index_type(self, idx):
    #assert isinstance(idx, Expr), \
    #    "Tuple index not an expression: %s" % idx
    #assert isinstance(idx, Const), "Unsupported expression: %s" % idx
    idx = int(idx.value)
    assert 0 <= idx < len(self.elt_types), \
        "Can't get element %d from tuple of length %d" % \
        (idx, len(self.elt_types))
    return self.elt_types[idx]

  def combine(self, other):
    if isinstance(other, TupleT) and \
       len(other.elt_types) == len(self.elt_types):
      combined_elt_types = [t1.combine(t2) for
                            (t1, t2) in zip(self.elt_types, other.elt_types)]
      if combined_elt_types != self.elt_types:
        return TupleT(combined_elt_types)
      else:
        return self
    else:
      raise IncompatibleTypes(self, other)

_tuple_types = {}
def repeat_tuple(t, n):
  """Given the base type t, construct the n-tuple t*t*...*t"""

  elt_types = tuple([t] * n)
  tuple_t = _tuple_types.get(elt_types)
  if tuple_t is not None:
    return tuple_t 
  tuple_t = TupleT(elt_types)
  _tuple_types[elt_types] = tuple_t
  return tuple_t

def make_tuple_type(elt_types):
  """
  Use this memoized construct to avoid constructing too many distinct tuple type
  objects and speeding up equality checks
  """

  key = tuple(elt_types)
  tuple_t = _tuple_types.get(key)
  if tuple_t is not None:
    return tuple_t 
  tuple_t = TupleT(key)
  _tuple_types[key] = tuple_t
  return tuple_t

empty_tuple_t = make_tuple_type(())



from .. ndtypes import ScalarT, ArrayT

from .. analysis.collect_vars import collect_var_names
from .. analysis import escape_analysis
from .. analysis.find_local_arrays import FindLocalArrays
from .. analysis.usedef import UseDefAnalysis
from .. syntax import AllocArray, ArrayView, Index, Struct, Var

from .. syntax import Adverb
from transform import Transform

class CopyElimination(Transform):
  def apply(self, fn):
    if all(isinstance(t, ScalarT) for t in fn.type_env.itervalues()):
      return fn
    else:
      return Transform.apply(self, fn)

  def pre_apply(self, fn):
    local_array_analysis = FindLocalArrays()
    local_array_analysis.visit_fn(fn)


    self.local_alloc = local_array_analysis.local_allocs
    self.local_arrays = local_array_analysis.local_arrays

    escape_info = escape_analysis(fn)
    self.may_escape = escape_info.may_escape
    self.may_alias = escape_info.may_alias
    self.may_return = escape_info.may_return
    
    self.usedef = UseDefAnalysis()
    self.usedef.visit_fn(fn)

    
    self.pointers_by_size = {}
    self.arrays_by_size = {}
    

  def no_array_aliases(self, array_name):
    alias_set = self.may_alias.get(array_name, [])
    array_aliases = [name for name in alias_set
                     if self.type_env[name].__class__ is ArrayT]
    # you're allowed one alias for yourself, but
    # any extras are other arrays with whom you share data
    # BEWARE: this will get convoluted and probably broken
    # if we ever have mutable compound objects in arrays
    return len(array_aliases) <= 1

  def is_array_alloc(self, expr):
    return expr.__class__ in [ArrayView, Struct, AllocArray] or isinstance(expr, Adverb)
    
  def transform_Assign(self, stmt):
    # pattern match only on statements of the form
    # dest[complex_indexing] = src
    # when:
    #   1) dest hasn't been used before as a value
    #   2) src doesn't escape
    #   3) src was locally allocated
    # ...then transform the code so instead of allocating src

    if stmt.lhs.__class__ is not Index or stmt.lhs.value.__class__ is not Var:
      return stmt 
    
    lhs_name = stmt.lhs.value.name

    # why assign to an array if it never gets used?
    if lhs_name not in self.usedef.first_use and \
       lhs_name not in self.may_escape and \
       self.no_array_aliases(lhs_name):
      return None
    
    # only match statements like array[idx] = some_rhs_var
    if stmt.lhs.type.__class__ is not ArrayT or stmt.rhs.__class__ is not Var:
      return stmt 
    
    curr_path = self.usedef.stmt_paths[id(stmt)]
    rhs_name = stmt.rhs.name

    
    if lhs_name in self.usedef.first_use and self.usedef.first_use[lhs_name] <= curr_path:
      return stmt 
    
    if self.usedef.last_use[rhs_name] != curr_path:
      return stmt
    
    if rhs_name in self.may_return:
      return stmt  
    
    if rhs_name not in self.local_arrays:
      return stmt 
            
    array_stmt = self.local_arrays[rhs_name]
    prev_path = self.usedef.stmt_paths[id(array_stmt)]
    if not self.is_array_alloc(array_stmt.rhs):
      return stmt 
    
    for lhs_depends_on in collect_var_names(stmt.lhs):
      created_on = self.usedef.created_on.get(lhs_depends_on)
      if created_on is None or created_on >= prev_path:
        return stmt 
             
    array_stmt.rhs = stmt.lhs
    return None
    


import numpy as np

from parakeet import jit
from parakeet.testing_helpers import run_local_tests, expect

int_mat = np.reshape(np.arange(6), (2,3))
float_mat = np.sqrt(int_mat)
bool_mat = int_mat % 2

matrices = [int_mat, float_mat, bool_mat]

@jit 
def mm_loops(X,Y,Z):
  m,d = X.shape
  n = Y.shape[1]
    
  for i in range(m):
      for j in range(n):
          total = 0
          for k in range(d):
              total += X[i,k] * Y[k,j]
          Z[i,j] = total 
  return Z

def test_matmult_loops():
  for X in matrices:
    for Y in matrices:
      res = np.dot(X, Y.T)
      Z = np.zeros(res.shape, dtype = res.dtype)
      expect(mm_loops, [X,Y.T,Z], res)
      
if __name__ == "__main__":
  run_local_tests()


from .. import names
from ..builder import build_fn
from ..syntax import Const, Var
from ..syntax.helpers import is_constant
from inline import Inliner
from transform import Transform

class SpecializeFnArgs(Transform):
  """
  Specialize the arguments to Map, Reduce, Scan, etc..  in two ways (including closure args):
    - constant arguments get moved into the function and are removed
      from the arguments list 
    - if an argument is repeated, only keep the first one. Be careful to not remove two instances
      of an array if it's being traversed along two different axes.
  
  TODO:
    - also get rid of unused arguments 
  """
  
  
  def build_arg_mapping(self, outer_closure_args, outer_array_args, inner_arg_names):
    """
    This function examines a list of actual args to a function (including its closure values), 
    and constructs:
      - a list of which closure and array args to keep
      - a list of which inner names to keep 
      - a dictionary associating discarded inner names with constants
      - a dictionary associating discarded inner names with kept names 
    """
    new_closure_args = []      
    remap = {}
    const = {}
    new_inner_names = []
      
    for pos, arg in enumerate(outer_closure_args):
      inner_name = inner_arg_names[pos]
      if arg.__class__ is Const:
        const[inner_name] = arg 
      elif arg in new_closure_args: 
        prev_pos = new_closure_args.index(arg)
        remap[inner_name] = new_inner_names[prev_pos]
      else:
        new_inner_names.append(inner_name)
        new_closure_args.append(arg)
      
    
    new_array_args = []
    for array_arg_pos, arg in enumerate(outer_array_args):
      pos = array_arg_pos + len(outer_closure_args)
      inner_name = inner_arg_names[pos]
      if arg.__class__ is Const:
        const[inner_name] = arg 
      elif arg in new_array_args:
        prev_pos = len(new_closure_args) + new_array_args.index(arg)
        remap[inner_name] = new_inner_names[prev_pos]
      else:
        new_inner_names.append(inner_name)
        new_array_args.append(arg)
        
    return new_closure_args, new_array_args, remap, const, new_inner_names

  def specialize_closure(self, clos, array_args):    
    closure_args = self.closure_elts(clos)
    fn = self.get_fn(clos)
    array_args = self.transform_expr_list(array_args)
    new_closure_args, new_array_args, remap, const, new_inner_names = \
      self.build_arg_mapping(closure_args, array_args, fn.arg_names)
    if len(new_inner_names) == len(fn.arg_names):
      return clos, array_args 
    
    new_input_types = [fn.type_env[name] for name in new_inner_names]
    
    new_name = "specialized_" + names.original(fn.name)
    new_fn, builder, _  = \
      build_fn(new_input_types,  
               return_type = fn.return_type, 
               name = new_name, 
               input_names = new_inner_names)
      
    new_fn.created_by = fn.created_by
     
    # call the original function 
    call_args = [] 
    for old_name, t  in zip(fn.arg_names, fn.input_types):
      if old_name in const:
        call_args.append(const[old_name])
      elif old_name in remap:
        other_name = remap[old_name]
        call_args.append(Var(name = other_name, type = t))
      else:
        assert old_name in new_inner_names, \
          "Expected %s to be in list of kept function arguments: %s" % (old_name, new_inner_names)
        call_args.append(Var(name = old_name, type = t))
    builder.return_(builder.call(fn, call_args))
    new_fn = Inliner().apply(new_fn)
    new_closure = self.closure(new_fn, new_closure_args)
    return new_closure, new_array_args

  def transform_Map(self, expr):
    if self.is_none(expr.axis) or is_constant(expr.axis):
      new_closure, new_array_args = self.specialize_closure(expr.fn, expr.args)
      expr.fn = new_closure
      expr.args = new_array_args
    return expr 
    
      


import numpy as np
import parakeet 

from parakeet.testing_helpers import expect, run_local_tests 

def test_ones_1d():
  expect(np.ones, [(10,)], np.ones(10))

def test_ones_2d():
  expect(np.ones, [(10,2)], np.ones((10,2)))

def test_ones_1d_int64():
  expect(np.ones, [(10,), np.int64], np.ones(10, dtype=np.int64))

def test_ones_2d_int64():
  expect(np.ones, [(10,2), np.int64], np.ones((10,2), dtype=np.int64))

def test_ones_1d_float64():
  expect(np.ones, [(10,), np.float64], np.ones(10, dtype=np.float64))

def test_ones_2d_float64():
  expect(np.ones, [(10,2), np.float64], np.ones((10,2), dtype=np.float64))

def test_ones_4d_float64():
  expect(np.ones, [(10,2,2,2), np.float64], np.ones((10,2,2,2), dtype=np.float64))

if __name__ == "__main__":
  run_local_tests()



from parakeet.testing_helpers import expect, run_local_tests

def test_min_int():
  expect(min, [-3, 4], -3)

def test_min_float():
  expect(min, [-3.0, 4.0], -3.0)


if __name__ == '__main__':
  run_local_tests()



import distutils
import platform  
import numpy.distutils as npdist 
import numpy.distutils.system_info as np_sysinfo 

import config 

config_vars = distutils.sysconfig.get_config_vars()

def get_compiler(_cache = {}):
  if config.compiler_path:
    return config.compiler_path
  if config.pure_c in _cache:
    return _cache[config.pure_c]
  for compiler in [('gcc' if config.pure_c else 'g++'), 
                   'icc']:
    path = distutils.spawn.find_executable(compiler)
    if path:
      _cache[config.pure_c] = path
      return path 
  assert False, "No compiler found!"
    

def get_source_extension():
  return ".c" if config.pure_c else ".cpp"

object_extension = ".o"
shared_extension = np_sysinfo.get_shared_lib_extension(True)


mac_os = platform.system() == 'Darwin'
windows = platform.system() == 'Windows'

python_include_dirs = [distutils.sysconfig.get_python_inc()]
numpy_include_dirs = npdist.misc_util.get_numpy_include_dirs()
include_dirs = numpy_include_dirs + python_include_dirs 

python_lib_dir = distutils.sysconfig.get_python_lib() + "/../../"
python_version = distutils.sysconfig.get_python_version()



from core_types import ConcreteT
from scalar_types import IntT 

###########################################
#
#  Pointers!
#
###########################################

class PtrT(ConcreteT):
  """
  I'm not giving pointer a concrete to_python/from_python conversion or any
  usable fields, so it's up to our ctypes_repr and llvm_backend to appropriately
  interpret objects of this type.
  """
  def __init__(self, elt_type):
    self.elt_type = elt_type 
    self._hash =  hash(elt_type)
    
  rank = 1
  def index_type(self, idx):
    assert isinstance(idx, IntT), \
        "Index into pointer must be of type int, got %s" % (idx)
    return self.elt_type

  def children(self):
    return self.elt_type

  def __str__(self):
    return "ptr(%s)" % self.elt_type

  def __eq__(self, other):
    return other.__class__ is PtrT and self.elt_type == other.elt_type

  def __hash__(self):
    return self._hash 

  def __repr__(self):
    return str(self)

  @property
  def ctypes_repr(self):
    return self._ctypes_repr

_ptr_types = {}
def ptr_type(t):
  if t in _ptr_types:
    return _ptr_types[t]
  else:
    ptr_t = PtrT(t)
    _ptr_types[t] = ptr_t
    return ptr_t




import numpy as np 

from parakeet import config 
from parakeet.testing_helpers import run_local_tests, expect 

config.print_transform_times = True 

w,h = 2,2
n = w *h 
vec = np.arange(n)
mat = vec.reshape((w,h))

dtypes = ['int8', 'int16', 'int32', 'int64', 
          'uint8', 'uint16', 'uint32', 'uint64', 
          'float32', 'float64', 'bool']

scalars = [True, 1, 1.0]
vecs = []
mats = []
for dtype_name in dtypes:
  scalars.append(np.dtype(dtype_name).type(3))
  vecs.append(vec.astype(dtype_name))
  mats.append(mat.astype(dtype_name))

def test_log1p_scalar():
  for scalar in scalars:
    print scalar, type(scalar)
    expect(np.log1p, [scalar], np.log1p(scalar))

def test_log1p_vec():
  for vec in vecs:
    expect(np.log1p, [vec], np.log1p(vec))

def test_log1p_mat():
  for mat in mats:
    expect(np.log1p, [mat], np.log1p(mat))

if __name__ == "__main__": 
    run_local_tests()




from .. import names, syntax 
from ..ndtypes import (IncompatibleTypes, 
                       Bool, Type,  ArrayT, Int64, TupleT,
                       NoneT, SliceT, ScalarT,  
                       make_tuple_type, make_array_type, lower_rank)
from ..syntax import (Assign, Tuple, Var, Cast, Return, Index, Map, 
                      ConstArrayLike, Const)
from ..syntax.helpers import get_types, zero_i64, none, const 
from ..transforms import Transform 


class RewriteTyped(Transform):
  def __init__(self, return_type = None):
    Transform.__init__(self, verify = False)
    self.forced_return_type = return_type 
    
    
  def pre_apply(self, fn):
    
    if self.forced_return_type is None:
      self.fn_return_type = self.fn.type_env["$return"]
    else:
      self.fn_return_type = self.forced_return_type
  
  def post_apply(self, fn):
    fn.return_type = self.fn_return_type  
  
  def coerce_expr(self, expr, t):
    assert t is not None
    expr = self.transform_expr(expr)
    if expr.type == t:
      return expr
    elif expr.__class__ is Tuple:
      if t.__class__ is not TupleT or \
          len(expr.type.elt_types) != t.elt_types:
        raise IncompatibleTypes(expr.type, t)
      else:
        new_elts = []
        for elt, elt_t in zip(expr.elts, t.elt_types):
          new_elts.append(self.coerce_expr(elt, elt_t))
        return Tuple(new_elts, type = t)
    else:
      assert \
          isinstance(expr.type, ScalarT) and \
          isinstance(t, ScalarT), \
          "Can't cast type %s into %s" % (expr.type, t)
      return syntax.Cast(expr, type=t)

  def transform_merge(self, merge):
    new_merge = {}
    for (var, (left, right)) in merge.iteritems():
      t = self.type_env[var]
      new_left = self.coerce_expr(left, t)
      new_right = self.coerce_expr(right, t)
      new_merge[var] = (new_left, new_right)
    return new_merge

  def transform_lhs(self, lhs):
    if isinstance(lhs, Var):
      t = self.lookup_type(lhs.name)
      if t == lhs.type:
        return lhs
      else:
        return Var(lhs.name, type = t)
    elif isinstance(lhs, syntax.Tuple):
      elts = map(self.transform_lhs, lhs.elts)
      elt_types = get_types(elts)
      if elt_types != lhs.type.elt_types:
        return syntax.Tuple(elts, type = make_tuple_type(elt_types))
      else:
        return lhs
    else:
      return lhs

  def transform_Var(self, expr):
    expr.type = self.fn.type_env[expr.name]
    return expr

  def transform_PrimCall(self, expr):
      arg_types = get_types(expr.args)
      upcast_types = expr.prim.expected_input_types(arg_types)
      result_type = expr.prim.result_type(upcast_types)
      upcast_args = [self.coerce_expr(x, t)
                     for (x,t) in zip(expr.args, upcast_types)]
      return syntax.PrimCall(expr.prim, upcast_args, type = result_type)

  def transform_Array(self, expr):

    array_t = expr.type
    elt_t = array_t.elt_type
    assert array_t.rank > 0
    if array_t.rank == 1:
      new_elts = [self.coerce_expr(elt, elt_t) for elt in expr.elts]
      return syntax.Array(new_elts, type = array_t)
    else:
      # need to allocate an output array and copy the elements in
      first_elt = self.assign_name(expr.elts[0], "first_elt")
      elt_dims = [self.shape(first_elt, i) for i in xrange(array_t.rank - 1)]
      n = len(expr.elts)
      outer_dim = const(n)
      all_dims = (outer_dim,) + tuple(elt_dims)
      array = self.alloc_array(elt_t, all_dims, "array_literal")
      for i, elt in enumerate(expr.elts):
        idx_expr = self.index(array, i, temp = False)
        # transform indexing to make missing indices explicit
        self.assign(idx_expr, expr.elts[i])
      return array

  def transform_Reduce(self, expr):
    acc_type = self.return_type(expr.combine)
    if expr.init and \
        not self.is_none(expr.init) and \
        expr.init.type != acc_type:
      assert len(expr.args) == 1
      expr.init = self.coerce_expr(expr.init, acc_type)
    return expr
    
  def transform_Scan(self, expr):
    acc_type = self.return_type(expr.combine)
    if expr.init and not self.is_none(expr.init) and expr.init.type != acc_type:
      if isinstance(acc_type, ScalarT):
        expr.init = self.coerce_expr(expr.init, acc_type)
      elif isinstance(expr.init.type, ScalarT) and \
           isinstance(expr.axis, Const) and \
           len(expr.args) == 1:
        arr_slice = self.slice_along_axis(expr.args[0], expr.axis, zero_i64)
        init_type = make_array_type(elt_type = expr.init.type, rank = arr_slice.type.rank)
        expr.init = ConstArrayLike(array = arr_slice, 
                                   value = expr.init, 
                                   type = init_type)
      else:
        assert False, \
          "Scan with scalar init of type %s and accumulator of type %s not yet supported" % \
                    (expr.init.type, acc_type)
    return expr

  def transform_Slice(self, expr):
    """
    # None step defaults to 1
    if isinstance(expr.step.type, core_types.NoneT):
      start_t = expr.start.type
      stop_t = expr.stop.type
      step = syntax_helpers.one_i64
      step_t = step.type
      slice_t = array_type.make_slice_type(start_t, stop_t, step_t)
      expr.step = step
      expr.type = slice_t
    """
    return expr
  

  def get_index_fn(self, array_t, index_types, _index_function_cache = {}):
    index_types = tuple(index_types)
    key = (array_t, index_types) 
    if key in _index_function_cache:
      return _index_function_cache[key]
    array_name = names.fresh("array")
    array_var = Var(array_name, type = array_t)
    idx_vars = []
    idx_names = []
    idx_types = []
    lower_rank_by = 0
    type_env = {array_name:array_t}
    
    for i, idx_t in enumerate(index_types):
      # indexing with None or a Slice doesn't decrease the rank
      # whereas by a scalar does 
      if isinstance(idx_t, ScalarT):
        lower_rank_by += 1 
      idx_name = names.fresh("idx%d" % (i+1))
      idx_names.append(idx_name)
      idx_var = Var(idx_name, type = idx_t)
      if idx_t is not Int64:
        idx_var = Cast(value = idx_var,  type = Int64)
        idx_t = Int64 
      idx_types.append(index_types)
      idx_vars.append(idx_var)
      type_env[idx_name] = idx_t
    
    elt_t = lower_rank(array_t, lower_rank_by)
    if len(idx_vars) > 1:
      idx = Tuple(idx_vars, type = make_tuple_type(index_types))
    else:
      idx = idx_vars[0]
    
    fn = syntax.TypedFn(
        name = names.fresh("fancy_indexing_helper"), 
        arg_names = (array_name,) + tuple(idx_names),
        input_types = (array_t,) + tuple(index_types),
        return_type = elt_t,
        type_env = type_env, 
        body = [Return (Index(array_var, idx, type = elt_t))]) 
    _index_function_cache[key] = fn 
    return fn 
    
  def transform_Index(self, expr):
    # TODO: Make fancy indexing work 
    # with multiple indices, boolean index elements, 
    # and multi-dimensional indexing
    
    index = expr.index
    if index.type.__class__ is TupleT:
      indices = self.tuple_elts(index) 
    else:
      indices = [index]
    
    if all(isinstance(idx.type, (NoneT, SliceT, ScalarT)) for idx in indices):
      return expr 
    
    map_args = []
    index_elt_types = []
    
    for index in indices:
      if index.type.__class__ is ArrayT:
        assert index.type.rank == 1, \
          "Don't yet support indexing by %s" % index.type 
        index_elt_t = index.type.elt_type
        
        if index_elt_t == Bool:
          assert False, "Indexing by boolean vector not yet implemented"
          #index_array = Where(expr.index)
          #index_elt_t = Int64
        else:
          map_args.append(expr.index)
          index_elt_types.append(index_elt_t)
      else:
        map_args.append(expr.index)
        index_elt_types.append(expr.index.type)
      
    index_fn = self.get_index_fn(expr.value.type, index_elt_types)
    index_closure = self.closure(index_fn, [expr.value])
    return Map(fn = index_closure, 
                   args = map_args, 
                   type = expr.value.type, 
                   axis = none)
    
  def transform_Assign(self, stmt):
    new_lhs = self.transform_lhs(stmt.lhs)
    lhs_t = new_lhs.type
    rhs = self.transform_expr(stmt.rhs)
    assert lhs_t is not None, "Expected a type for %s!" % stmt.lhs
    if new_lhs.__class__ is Tuple: 
      rhs = self.assign_name(rhs)
      for (i, lhs_elt) in enumerate(new_lhs.elts):
        if lhs_elt.__class__ is Var:
          name = names.original(lhs_elt.name) 
        else:
          name = None 
        idx = self.index(rhs, i, name = name )
        elt_stmt = self.transform_Assign(Assign(lhs_elt, idx))
        self.blocks.append_to_current(elt_stmt)
      return None  
    elif new_lhs.__class__ is Index and \
         isinstance(new_lhs.value.type, ArrayT) and \
         isinstance(rhs.type, ScalarT):
      new_rhs = self.coerce_expr(rhs, new_lhs.value.type.elt_type)
      assert new_rhs.type and isinstance(new_rhs.type, Type), \
          "Expected type annotation on %s, but got %s" % (new_rhs, new_rhs.type)
      
      stmt.lhs = new_lhs
      stmt.rhs = new_rhs
      return stmt 

    else:     
      new_rhs = self.coerce_expr(rhs, lhs_t)
      assert new_rhs.type and isinstance(new_rhs.type, Type), \
          "Expected type annotation on %s, but got %s" % (new_rhs, new_rhs.type)
      stmt.lhs = new_lhs
      stmt.rhs = new_rhs
      return stmt
    
  def transform_If(self, stmt):
    stmt.cond = self.coerce_expr(stmt.cond, Bool)
    stmt.true = self.transform_block(stmt.true)
    stmt.false = self.transform_block(stmt.false)
    stmt.merge = self.transform_merge(stmt.merge)
    return stmt

  def transform_Return(self, stmt):
    stmt.value = self.coerce_expr(stmt.value, self.fn_return_type)
    return stmt

  def transform_ForLoop(self, stmt):
    var_t = stmt.var.type
    stmt.start = self.coerce_expr(stmt.start, var_t)
    stmt.stop = self.coerce_expr(stmt.stop, var_t)
    stmt.step = self.coerce_expr(stmt.step, var_t)
    stmt.body = self.transform_block(stmt.body)
    stmt.merge = self.transform_merge(stmt.merge)
    return stmt 
  
  def transform_While(self, stmt):
    stmt.cond = self.coerce_expr(stmt.cond, Bool)
    stmt.body = self.transform_block(stmt.body)
    stmt.merge = self.transform_merge(stmt.merge)
    return stmt

def rewrite_typed(typed_fundef, return_type = None):
  return RewriteTyped(return_type).apply(typed_fundef)



from parakeet.ndtypes import Int8, Int16, Int32, Float32, Float64
from parakeet.ndtypes import is_scalar_subtype
from parakeet.testing_helpers import run_local_tests 

def test_int32_subtypes():
  assert is_scalar_subtype(Int8, Int32)
  assert not is_scalar_subtype(Int32, Int8)
  assert is_scalar_subtype(Int32, Float32)
  assert not is_scalar_subtype(Float32, Int32)

def test_float32_subtypes():
  assert is_scalar_subtype(Int8, Float32)
  assert not is_scalar_subtype(Float32, Int8)
  assert is_scalar_subtype(Float32, Float64)
  assert not is_scalar_subtype(Float64, Float32)

if __name__ == '__main__':
  run_local_tests()



from .. import names, prims  
from ..ndtypes import ScalarT, Type, type_conv  
from .. syntax import  FormalArgs, Var, UntypedFn, Return, PrimCall, Expr, Cast 


_untyped_fn_cache = {}
def simple_untyped_fn(name, 
                        expr,
                        n_inputs = 1, 
                        fixed_args  = [], 
                        keyword_args = {},                         
                        unpack = False):
  key = name, expr, n_inputs, tuple(fixed_args), tuple(keyword_args.items()), unpack
  if key in _untyped_fn_cache:
    return _untyped_fn_cache[key]
  
  fn_name = names.fresh(name)
  args_obj = FormalArgs()

  arg_vars = []
  for name in names.fresh_list(n_inputs):
    args_obj.add_positional(name)
    arg_vars.append(Var(name))
  
  if unpack:
    combined_args = tuple(fixed_args) + tuple(arg_vars)
  else:
    combined_args = tuple(fixed_args) + (tuple(arg_vars),)
  result = expr(*combined_args, **keyword_args)
  body = [Return(result)]
  fundef = UntypedFn(fn_name, args_obj, body, [])
  _untyped_fn_cache[key] = fundef 
  return fundef 
  
def build_untyped_prim_fn(p):
  """Given a primitive, return an untyped function which calls that prim"""
  assert isinstance(p, prims.Prim), "Expected Prim but got %s" % p 
  return simple_untyped_fn(p.name, PrimCall, p.nin, [p])


def build_untyped_expr_fn(expr, n_args = 1):
  """Given an expression, return a function which applies that expression to arguments"""
  return simple_untyped_fn(expr.__name__ + "_fn", expr, n_args)
    
_untyped_cast_wrappers = {}
def build_untyped_cast_fn(t):
  if not isinstance(t, Type):
    t = type_conv.equiv_type(t)
  assert isinstance(t, ScalarT), "Expected scalar type but got %s" % t 
  return simple_untyped_fn("cast_" + str(t), Cast, 1, keyword_args = {'type': t})
  


from expr import Expr 

class SeqExpr(Expr):
  def __init__(self, value, type = None, source_info = None):
    self.value = value 
    self.type = type 
    self.source_info = source_info 

  def children(self):
    yield self.value 

class Enumerate(SeqExpr):
  pass 
  
class Zip(SeqExpr):
  def __init__(self, values, type = None, source_info = None):
    self.values = tuple(values) 
    self.type = type 
    self.source_info = source_info
    
  def children(self):
    return self.values

class Len(SeqExpr):
  pass 
  
class Index(SeqExpr):
  """
  TODO: 
    - make all user-defined indexing check_negative=True by default 
    - implement backend logic for lowering check_negative 
  """
  def __init__(self, value, index, check_negative = None, type = None, source_info = None):
    self.value = value 
    self.index = index 
    self.check_negative = check_negative 
    self.type = type 
    self.source_info = source_info
  
  def __eq__(self, other):
    return other.__class__ is Index and \
           other.value == self.value and \
           other.index == self.index
  
  def __hash__(self):
    return hash((self.value, self.index))
  
  def children(self):
    yield self.value
    yield self.index

  def __str__(self):
    return "%s[%s]" % (self.value, self.index)
  



import numpy as np 



from parakeet.analysis import mutability_analysis
from parakeet import make_array_type, Int64, Bool, ptr_type, specialize
from parakeet.transforms.pipeline import lowering 
from parakeet.testing_helpers import run_local_tests

def f(x,y):
    x[0] = True
    return y

def test_mutable_array():
  bool_vec = np.array([True, False])
  int_vec = np.array([1,2])
  typed, _, =  specialize(f, [bool_vec, int_vec])
   
  mutable_types = mutability_analysis.find_mutable_types(typed)
  int_array_t = make_array_type(Int64, 1)
  bool_array_t = make_array_type(Bool, 1)
  assert int_array_t not in mutable_types, \
      "Didn't expect %s in mutable_types %s" % (int_array_t, mutable_types)
  assert bool_array_t in mutable_types, \
      "Expected %s in mutable_types %s" % (bool_array_t, mutable_types)
      
  lowered = lowering.apply(typed)
  mutable_types = mutability_analysis.find_mutable_types(lowered)
  ptr_bool_t = ptr_type(Bool)
  ptr_int_t = ptr_type(Int64) 
  assert ptr_int_t not in mutable_types, \
      "Didn't expect %s in lowered mutable types %s" % \
      (ptr_int_t, mutable_types)
  assert ptr_bool_t in mutable_types, \
      "Expected %s in lowered mutable_types %s" % (ptr_bool_t, mutable_types)

"""
# Removed this test when I made attributes immutable 
def f():
  x = slice(1, 2, 3)
  x.start = 10
  x.stop = 20
  y = slice(1,2,3)
  if 3 < y.step:
    y.step = y.step + 1
  else:
    y.step = 0

def test_mutable_slice():
  _, typed, _, _ =  testing_helpers.specialize_and_compile(f, [])
  mutable_types = mutability_analysis.find_mutable_types(typed)

  assert len(mutable_types) == 1, mutable_types
  lowered = lowering.apply(typed)
  mutable_types = mutability_analysis.find_mutable_types(lowered)
  assert len(mutable_types) == 1, mutable_types
"""
if __name__ == '__main__':
  run_local_tests()




from expr import Expr 


class DelayUntilTyped(Expr):
  """
  Once the list of values has been annotated with locally inferred types, 
  pass them to the given function to construct a final expression 
  """
  def __init__(self, values, keywords, fn, source_info = None):
    """
    No need for a 'type' argument since the user-supplied function 
    will generate a typed expression to replace this one
    """
    if isinstance(values, list): values = tuple(values)
    elif not isinstance(values, tuple): values = (self.values,)
    self.values = values 
    
    if keywords is None: keywords = {}
    self.keywords = keywords 
    
    self.fn = fn 
    self.source_info = source_info 
    self.type = None 
    
  def children(self):
    return tuple(self.values) + tuple(self.keywords.values())  


from parakeet.testing_helpers import run_local_tests, expect_eq

import parakeet


def identity_i64(x):
  fn = parakeet.mk_identity_fn(parakeet.Int64)
  print repr(fn) 
  return parakeet.run_typed_fn(fn, (x,))
  
def identity_f64(x):
  fn = parakeet.mk_identity_fn(parakeet.Float64)
  return parakeet.run_typed_fn(fn, (x,))

def test_identity():
  expect_eq(identity_i64(1), 1)
  expect_eq(identity_i64(-1), -1)
  expect_eq(identity_f64(1.0), 1.0)
  expect_eq(identity_f64(-1.0), -1.0)


if __name__ == '__main__':
 
  run_local_tests()  



from parakeet import jit 

def covariance(x,y):
  return ((x-x.mean()) * (y-y.mean())).mean()

@jit 
def fit_simple_regression(x,y):
  slope = covariance(x,y) / covariance(x,x)
  offset = y.mean() - slope * x.mean() 
  return slope, offset

import numpy as np 

N = 10**4
x = np.random.randn(N).astype('float64')
slope = 903.29
offset = 102.1
y = slope * x + offset

slope_estimate, offset_estimate = fit_simple_regression(x,y)

print "Expected slope =", slope, "offset =", offset
print "Parakeet slope =", slope_estimate, "offset =", offset_estimate



import numpy as np


from parakeet.testing_helpers import run_local_tests, expect

mat_shape = (2,3)
vec_len = np.prod(mat_shape)
int64_vec = np.arange(vec_len)

def get_vector(t):
  if t == 'bool':
    return int64_vec % 2
  else:
    return int64_vec.astype(t)

def run_vv(xtype, ytype):
  x = get_vector(xtype)
  y = get_vector(ytype)
  expect(np.dot, [x,y], np.dot(x,y))


def run_mv(xtype, ytype):
  x = get_vector(xtype)
  y = get_vector(ytype)
  x = x.reshape(mat_shape)
  y = y[:x.shape[1]]
  expect(np.dot, [x,y], np.dot(x,y))


def run_vm(xtype, ytype):
  x = get_vector(xtype)
  y = get_vector(ytype)
  
  y = y.reshape(mat_shape)
  x = x[:y.shape[0]]

  expect(np.dot, [x,y], np.dot(x,y))

def run_mm(xtype, ytype):
  x = get_vector(xtype)
  y = get_vector(ytype)
  x = x.reshape(mat_shape)
  y = y.reshape(mat_shape).T
  expect(np.dot, [x,y], np.dot(x,y))

#
# Vector-Vector
#
def test_dot_vv_i64_i32():
  run_vv('int64', 'int32')

def test_dot_vv_i64_f32():
  run_vv('int64', 'float32')

def test_dot_vv_i64_bool():
  run_vv('int64', 'bool')

def test_dot_vv_f64_f32():
  run_vv('float64', 'float32')

def test_dot_vv_f64_i32():
  run_vv('float64', 'int32')

def test_dot_vv_f64_bool():
  run_vv('float64', 'bool')

def test_dot_vv_i32_i32():
  run_vv('int32', 'int32')

def test_dot_vv_f32_f32():
  run_vv('float32', 'float32')

def test_dot_vv_bool_bool():
  run_vv('bool', 'bool')
#
# Matrix-Matrix
#

def test_dot_mm_i64_i32():
  run_mm('int64', 'int32')

def test_dot_mm_i64_f32():
  run_mm('int64', 'float32')

def test_dot_mm_bool_f32():
  run_mm('bool', 'float32')

#
#  Matrix-Vector 
#


def test_dot_mv_i64_i32():
  run_mv('int64', 'int32')

def test_dot_mv_i64_i32():
  run_mv('int64', 'int32')

#
# Vector-Matrix
#
def test_dot_vm_i64_i32():
  run_vm('int64', 'int32')

def test_dot_vm_bool_float64():
  run_vm('bool', 'float64')

if __name__ == '__main__':
    run_local_tests()




from ..ndtypes import make_closure_type
from expr import Expr
from formal_args import FormalArgs 
from stmt import block_to_str


class UntypedFn(Expr):
  """
  Function definition.
  A top-level function can have references to python values from its enclosing
  scope, which are stored in the 'python_refs' field.

  A nested function, on the other hand, might refer to some variables from its
  enclosing Parakeet scope, whose original names are stored in
  'parakeet_nonlocals'
  """
    
  registry = {}

  def __init__(self, name, args, body, 
               python_refs = None, 
               parakeet_nonlocals = None, 
               doc_string = None, 
               source_info = None):
    assert isinstance(name, str), "Expected string for fn name, got %s" % name
    self.name = name 
    
    assert isinstance(args, FormalArgs), \
        "Expected arguments to fn to be FormalArgs object, got %s" % self.args
    self.args = args 
    
    assert isinstance(body, list), \
        "Expected body of fn to be list of statements, got " + str(body)
    self.body = body 
    
    self.python_refs = python_refs 
    self.parakeet_nonlocals = parakeet_nonlocals 
    
    self.type = make_closure_type(self, ())
    
    self.source_info = source_info 
    self.registry[self.name] = self

    
  
  def __repr__(self):
    return "def %s(%s):%s" % (self.name, self.args, block_to_str(self.body))
  
  def __str__(self):
    return repr(self)
  
  def __hash__(self):
    return hash(self.name)

  def python_nonlocals(self):
    if self.python_refs:
      return [ref.deref() for ref in self.python_refs]
    else:
      return []

  def children(self):
    return ()



from .. syntax import Var, Tuple, Index, Attribute  
from syntax_visitor import SyntaxVisitor

class StmtPath(object):
  def __init__(self, elts = ()):
    self.elts = tuple(elts) 
  
  def __str__(self):
    return "{%s}" % ", ".join(str(e) for e in self.elts)
  
  def __len__(self):
    return len(self.elts)

  def __getitem__(self, i):
    return self.elts[i]
  
  def __iter__(self):
    return iter(self.elts)
  
  def __lt__(self, other):
    if other.__class__ is not StmtPath:
      other = StmtPath([other])
    n = len(other)
    for (i,c1) in enumerate(self.elts):
      if n <= i:
        return False 
      c2 = other[i]
      if type(c1) is int and type(c2) is int:
        if c1 < c2:
          return True 
        elif c1 > c2:
          return False 
      # if you're not on the same branch of control flow
      else:
        assert type(c1) is bool and type(c2) is bool 
        if c1 != c2:
          return False
    return False
  
  
  def __eq__(self, other):
    if other.__class__ is not StmtPath:
      other = StmtPath([other])
      
    if len(self) != len(other):
      return False
    else:
      for (i,c) in enumerate(self.elts):
        if c != other[i]:
          return False 
      return True 
  
  def __ne__(self, other):
    return not (self == other)
  
  def __le__(self, other):
    return self == other or self < other
  
  def __lte__(self, other):
    return (self < other) or (self == other)
  
  def __gt__(self, other):
    return not (self <= other) 
  
  def __gte__(self, other):
    return not (self < other)

class UseDefAnalysis(SyntaxVisitor):
  """
  Number all the statements and track the creation as well as first and last
  uses of variables
  """
  def __init__(self):
    # map from pointers of statement objects to
    # nested sequential numbering
    self.stmt_paths = {}

    # statement count per scope 
    self.stmt_counters = [0]
    
    # the scopes leading up to the current one 
    self.path_prefix = []
    
    
    # map from variable names to counter number of their
    # first/last usages
    self.first_use = {}
    self.last_use = {}
    self.first_read = {}
    self.last_read = {}
    self.first_write = {}
    self.last_write = {}
    self.created_on = {}
  
  def counter(self):
    return self.stmt_counters[-1]
  
  def inc_counter(self):
    self.stmt_counters[-1] += 1
  
  def push_counter(self):
    self.stmt_counters.append(0)
    
  def pop_counter(self):
    return self.stmt_counters.pop()
  
  def push_scope(self, c = None):
    if c is None:
      c = self.counter()
    self.path_prefix.append(c) 
  
  def pop_scope(self):
    self.path_prefix.pop()
  
  def curr_path(self):
    return StmtPath(self.path_prefix + [self.counter()])

  def precedes_stmt_id(self, stmt_id1, stmt_id2):
    return self.stmt_paths[stmt_id1] < self.stmt_paths[stmt_id2] 
    
  def precedes_stmt(self, stmt1, stmt2):
    return self.stmt_paths[id(stmt1)] < self.stmt_paths[id(stmt2)] 
    
  def visit_fn(self, fn):
    for name in fn.arg_names:
      self.created_on[name] = 0
    SyntaxVisitor.visit_fn(self, fn)

  def visit_lhs(self, expr):
    if expr.__class__ is Var:
      self.created_on[expr.name] = self.curr_path()
    elif expr.__class__ is Tuple:
      for elt in expr.elts:
        self.visit_lhs(elt)
    else:
      assert expr.__class__ in (Attribute, Index)
      assert expr.value.__class__ is Var
      value_name = expr.value.name 
      p = self.curr_path()
      if value_name not in self.first_write:
        self.first_write[value_name] = p 
      self.last_write[value_name] = p

  def vist_merge(self, merge):
    p = self.curr_path()
    for (name, (l,r)) in merge.iteritems():
      self.visit_expr(l)
      self.visit_expr(r)
      self.created_on[name] = p
  
  def visit_Var(self, expr):
    name = expr.name
    p = self.curr_path()
    if name not in self.first_use:
      self.first_use[name] = p
    self.last_use[name]= p

  def visit_block(self, stmts, branch_label = None):
    self.push_scope()
    if branch_label is not None:
      self.push_scope(branch_label)
    SyntaxVisitor.visit_block(self, stmts)
    if branch_label is not None:
      self.pop_scope()
    self.pop_scope()
    
  def visit_If(self, stmt):
    self.visit_expr(stmt.cond)
    self.visit_block(stmt.true, True)
    self.visit_block(stmt.false, False)
    self.visit_merge(stmt.merge)
    
  def visit_stmt(self, stmt):
    self.inc_counter()
    self.stmt_paths[id(stmt)] = self.curr_path()
    SyntaxVisitor.visit_stmt(self, stmt)



import ctypes
from core_types import Type, IncompatibleTypes, StructT, ImmutableT
from scalar_types import  Int64
import type_conv 

###########################################
#
#  Closures!
#
###########################################

class ClosureT(StructT, ImmutableT):
  def __init__(self, fn, arg_types):
    
    self.fn = fn 
    
    if arg_types is None:
      arg_types = ()
    elif not hasattr(arg_types, '__iter__'):
      arg_types = tuple([arg_types])
    else:
      arg_types = tuple(arg_types)
      
    self.arg_types = arg_types 
    self._hash = hash( (fn,) + arg_types)
  
    self.specializations = {}

  def children(self):
    return self.arg_types
  
  def __hash__(self):
    return self._hash 

  def __eq__(self, other):
    
    if other.__class__ is not ClosureT:
      return False
    if self.fn.__class__ != other.fn.__class__:
      return False
    if isinstance(self.fn, str):
      same_fn = (self.fn == other.fn)
    else:
      # checking equality of functions isn't really defined, 
      # so just check that it's the same location 
      same_fn = self.fn is other.fn
    return same_fn and self.arg_types == other.arg_types

  def __str__(self):
    fn_name = self.fn if isinstance(self.fn, str) else self.fn.name
    return "ClosT(%s, {%s})" % (fn_name, ", ".join(str(t)
                                                   for t in self.arg_types))




  def combine(self, other):
    if isinstance(other, ClosureSet):
      return other.combine(self)
    elif isinstance(other, ClosureT):
      if self == other:
        return self
      else:
        return ClosureSet(self, other)
    else:
      raise IncompatibleTypes(self, other)

_closure_type_cache = {}
def make_closure_type(fn, closure_arg_types = []):
  closure_arg_types = tuple(closure_arg_types)
  key = (fn, closure_arg_types)
  if key in _closure_type_cache:
    return _closure_type_cache[key]
  else:
    t = ClosureT(fn, closure_arg_types)
    _closure_type_cache[key] = t
    return t





class ClosureSet(Type):
  """
  If multiple closures meet along control flow paths then join them into a
  closure set. This type should not appear by the time we're generating LLVM
  code.
  """

  _members = ['closures']

  def __init__(self, *closures):
    self.closures = set([])
    for clos_t in closures:
      if isinstance(clos_t, ClosureSet):
        self.closures.update(clos_t.closures)
      else:
        assert isinstance(clos_t, ClosureT)
        self.closures.add(clos_t)

  def combine(self, other):
    if isinstance(other, ClosureSet):
      combined_closures = self.closures.union(other.closures)
      if combined_closures != self.closures:
        return ClosureSet(combined_closures)
      else:
        return self
    else:
      raise IncompatibleTypes(self, other)

  def __eq__(self, other):
    return self.closures == other.closures

  def __iter__(self):
    return iter(self.closures)

  def __len__(self):
    return len(self.closures)



import numpy as np
import time

from parakeet import allpairs, multiply 
from parakeet.testing_helpers import expect, expect_allpairs, run_local_tests

bool_vec = np.array([True, False, ])
int_vec = np.array([1,2,])
float_vec = np.array([10.0, 20.0, ])

vectors = [bool_vec, int_vec, float_vec]

def loop_outer_prod(x,y,z):
  nx = x.shape[0]
  ny = y.shape[0]
  for i in xrange(0, nx):
    for j in xrange(0, ny): 
      z[i,j] = x[i] * y[j]
  return z

def test_loop_outer_prod():
  for v1 in vectors:
    for v2 in vectors:
      res = np.outer(v1, v2)
      v3 = np.zeros_like(res)
      expect(loop_outer_prod, [v1, v2, v3], res)

def adverb_outer_prod(x,y):
  return allpairs(multiply, x, y)

def test_adverb_outer_prod():
  expect_allpairs(adverb_outer_prod, np.multiply.outer, vectors)

if __name__ == '__main__':
  run_local_tests()



import math  
import numpy as np
import lib, prims   
from syntax import Zip, Len


function_mappings = {
  # PYTHON BUILTINS          
  zip : Zip, 
  map : lib.map, 
  reduce : lib.reduce, 
  tuple : lib.builtin_tuple, 
  range : lib.arange, 
  xrange : lib.arange, 
  float : lib.numpy_types.float64, 
  int : lib.numpy_types.int64, 
  long : lib.numpy_types.int64, 
  bool : lib.numpy_types.bool, 
  len : Len,  
  min : lib.builtin_min,
  max : lib.builtin_max,
  all : lib.reduce_all, 
  any : lib.reduce_any, 
  sum : lib.reduce_sum, 
  abs : prims.abs,  
  pow : prims.power, 
  
  # TYPES 
  np.int8 : lib.numpy_types.int8, 
  np.int16 : lib.numpy_types.int16, 
  np.int32 : lib.numpy_types.int32, 
  np.int64 : lib.numpy_types.int64, 
  np.uint8 : lib.numpy_types.uint8,
  np.uint16 : lib.numpy_types.uint16, 
  np.uint32 : lib.numpy_types.uint32, 
  np.uint64 : lib.numpy_types.uint64,  
  np.bool : lib.numpy_types.bool, 
  np.bool8 : lib.numpy_types.bool, 
  np.bool_ : lib.numpy_types.bool, 
  np.float32 : lib.numpy_types.float32, 
  np.float64 : lib.numpy_types.float64, 
  
  np.rank : lib.rank, 
  np.alen : Len, 
  np.real : lib.real, 
  # np.imag : lib.imag
  np.size : lib.size, 
  
  np.minimum : prims.minimum, 
  np.maximum : prims.maximum, 
  
  # COMPARISONS
  np.greater : prims.greater, 
  np.greater_equal : prims.greater_equal, 
  np.equal : prims.equal, 
  np.not_equal : prims.not_equal, 
  np.less : prims.less, 
  np.less_equal : prims.less_equal, 
  
  # REDUCTIONS 
  np.min : lib.reduce_min, 
  np.max : lib.reduce_max,
  
  np.argmin : lib.argmin, 
  np.argmax : lib.argmax, 
  
  np.all : lib.reduce_all, 
  
  np.any : lib.reduce_any, 
  np.sum : lib.reduce_sum, 
  np.prod : lib.prod, 
  np.mean : lib.mean, 
  
  
  np.abs : prims.abs, 

  # ARRAY CONSTRUCTORS 
  np.array : lib.array, 
  np.tile : lib.tile, 
  # np.eye : , 
  np.arange  : lib.arange, 
  
  np.empty_like : lib.empty_like, 
  np.empty : lib.empty, 
  np.zeros_like : lib.zeros_like, 
  np.zeros : lib.zeros, 
  np.ones : lib.ones, 
  np.ones_like : lib.ones_like, 
  
  # ARITHMETIC 
  np.add : prims.add, 
  np.subtract : prims.subtract, 
  np.multiply : prims.multiply,
   
  np.divide : prims.divide, 
  np.floor_divide : lib.floor_divide, 
  np.true_divide : lib.true_divide, 
   
  np.mod : prims.remainder,
  np.remainder : prims.remainder,
  np.fmod : prims.fmod,
  # np.modf : prims.modf,
  
  np.logical_and : prims.logical_and, 
  np.logical_not : prims.logical_not, 
  np.logical_or : prims.logical_or,  
    
  math.sqrt : prims.sqrt, 
  np.sqrt : prims.sqrt, 
  
  np.sign : lib.sign, 
  np.reciprocal : lib.reciprocal, 
  np.conjugate : lib.conjugate,
  
  # ROUNDING
  
  np.trunc : prims.trunc, 
  math.trunc : prims.trunc,
   
  np.rint : prims.rint,
  
  np.floor : prims.floor, 
  math.floor : prims.floor,
   
  np.ceil : prims.ceil,
  math.ceil : prims.ceil, 
  
  np.round : prims.round,
 
  # LOGS AND EXPONENTIATION 
  np.square : lib.square,
  math.pow : prims.power, 
  np.power : prims.power,
  
  np.exp : prims.exp,
  math.exp : prims.exp, 
   
  np.exp2 : prims.exp2,
   
  np.expm1 : prims.expm1,
  math.expm1 : prims.expm1, 
   
  np.log : prims.log, 
  math.log : prims.log, 
  
  np.log10 : prims.log10,
  math.log10 : prims.log10, 
   
  np.log2 : prims.log2,
   
  np.log1p : prims.log1p,
  math.log1p : prims.log1p, 
  
  np.logaddexp : lib.logaddexp, 
  np.logaddexp2 : lib.logaddexp2, 
  
  # TRIG 
  np.cos : prims.cos, 
  math.cos : prims.cos, 
  np.sin : prims.sin, 
  math.sin : prims.sin, 
  np.tan : prims.tan, 
  math.tan : prims.tan, 
  np.arccos : prims.arccos, 
  math.acos: prims.arccos, 
  np.arcsin : prims.arcsin, 
  math.asin : prims.arcsin, 
  np.arctan : prims.arctan, 
  math.atan : prims.arctan, 
  np.arctan2 : prims.arctan2, 
  math.atan2 : prims.arctan2, 
  math.cosh : prims.cosh, 
  np.cosh : prims.cosh, 
  np.sinh : prims.sinh,           
  math.sinh : prims.sinh, 
  np.tanh : prims.tanh, 
  math.tanh : prims.tanh, 
  math.atanh : prims.arctanh,
  np.arctanh : prims.arctanh,  
  math.asinh : prims.arcsinh, 
  np.arcsinh : prims.arcsinh, 
  math.acosh : prims.arccosh, 
  np.arccosh : prims.arccosh,
  np.rad2deg : lib.rad2deg, 
  np.deg2rad : lib.deg2rad,  
  # np.hypot : lib.hypot,
  
  np.where : lib.where,
  np.linspace : lib.linspace, 
  np.vdot : lib.vdot, 
  np.dot : lib.dot,
  np.linalg.norm : lib.linalg.norm,  
}

property_mappings = {
  'dtype' : lib.get_elt_type,                
  # 'imag' : lib.imag,      
  'itemsize' : lib.itemsize,
  'real' :  lib.identity, # ain't no complex numbers yet 
  'shape' : lib.shape, 
  'size' : lib.size,   
  # 'strides' : lib.strides, 
  'ndim' : lib.rank, 
  'T' : lib.transpose, 
}


method_mappings = {
  'fill' : lib.fill, 
  'any' : lib.reduce_any, 
  'all' : lib.reduce_all, 
  'argmax' : lib.argmax, 
  # 'argsort' : lib.argsort, 
  'copy' : lib.copy, 
  'cumprod' : lib.cumprod, 
  'cumsum' : lib.cumsum, 
  'dot' : lib.dot, 
  'fill' : lib.fill, 
  'flatten' : lib.ravel, 
  # 'diagonal' : lib.diagonal, 
  
  'mean' : lib.mean, 
  
  'min' : lib.reduce_min,
  'max' : lib.reduce_max,
  
  'ravel' : lib.ravel, 
  'transpose' : lib.transpose,
  'sum' : lib.reduce_sum,
  }



from parakeet import testing_helpers
import numpy as np 

def ravel_method(x):
  return x.ravel()

vec  = np.arange(12)

def run(x):
  testing_helpers.expect(ravel_method, [x], x.ravel())
  # testing_helpers.expect(np.ravel, [x], x.ravel())

def test_ravel_vec():
  run(vec)

def test_ravel_mat_row_major():
  matrix_row_major = np.reshape(vec, (3,4), order = "C")
  run(matrix_row_major)

def test_ravel_mat_col_major():
  matrix_col_major = np.reshape(vec, (3,4), order = "F")
  run(matrix_col_major)

def test_ravel_cube():
  cube = np.reshape(vec, (3,2,2))
  run(cube)

if __name__ == '__main__':
    testing_helpers.run_local_tests()



from parakeet import jit 
from parakeet.testing_helpers import run_local_tests, expect 
import time 

def f(x,y,z=1,q=3):
  return x + y + z + q


jitf = jit(f)

def test_call_overhead_identity():
  n = 2000
  x = 3 
  start_t = time.time()
  for i in xrange(n):
    f(x,x,x,x)
  python_time = time.time() - start_t 
  print "Python time for %d calls: %f" % (n, python_time)
  # warm up!
  jitf(x,x,x,x)
  start_t = time.time()
  for i in xrange(n):
    jitf(x,x,x,x)
  parakeet_time = time.time() - start_t
  print "Parakeet time for %d calls: %f" % (n, parakeet_time)
  print "Slowdown = %f" % (parakeet_time / python_time )
  assert parakeet_time < 1000 * python_time, "Excessive call overhead: %f Python vs. %f Parakeet" % (python_time, parakeet_time)


if __name__ == "__main__":
  run_local_tests()
  

  



import numpy as np 
from parakeet import jit, testing_helpers 

@jit 
def true_divided(x):
    return True / x

def test_true_divided_bool():
    testing_helpers.expect(true_divided, [True], True)

def test_true_divided_int():
    testing_helpers.expect(true_divided, [1], 1)
    testing_helpers.expect(true_divided, [2], 0)

def test_true_divided_float():
    testing_helpers.expect(true_divided, [1.0], 1.0)
    testing_helpers.expect(true_divided, [2.0], 0.5)

def test_true_divided_uint8():
    testing_helpers.expect(true_divided, [np.uint8(1)], 1)
    testing_helpers.expect(true_divided, [np.uint8(2)], 0)

if __name__ == '__main__':
    testing_helpers.run_local_tests()




import numpy as np 

from .. ndtypes import ArrayT, TupleT   
from .. syntax import TypedFn, Var
from ..analysis import SyntaxVisitor
from abstract_value import one, zero, Const, unknown, Tuple, Array, Struct, abstract_tuple  





def bind_list(arg_names, abstract_values):
  assert len(arg_names) == len(abstract_values)
  env = {}
  for name, abstract_value in zip(arg_names, abstract_values):
    env[name] = abstract_value
  return env 

_cache = {}
def symbolic_call(fn, symbolic_inputs):
  key = fn.cache_key, tuple(symbolic_inputs)
  if key in _cache:
    return _cache[key]
  analysis = FindConstantValues(fn, symbolic_inputs)
  return_value = analysis.visit_fn(fn)
  result = (analysis.env, return_value)
  _cache[key] = result 
  return result 

class FindConstantValues(SyntaxVisitor):
  def __init__(self, fn, abstract_values):
    self.env = bind_list(fn.arg_names, abstract_values)
    self.return_value = None 
    
  def visit_expr(self, expr):
    result = SyntaxVisitor.visit_expr(self, expr)
    if result is None: 
      return unknown
    else: 
      return result 
  
  def visit_fn(self, fn):
    SyntaxVisitor.visit_fn(self, fn)
    
    if self.return_value is None:
      return unknown 
    else:
      return self.return_value
  
  def visit_Call(self, expr):
    if expr.fn.__class__ is TypedFn: 
      args = self.visit_expr_list(expr.args)
      _, return_value = symbolic_call(expr.fn, args)
      return return_value 
    else:
      return unknown 

  def visit_Var(self, expr):
    return self.env.get(expr.name, unknown)
  
  def visit_Const(self, expr):
    if expr.value == 0:
      return zero
    elif expr.value == 1:
      return one 
    elif isinstance(expr.value, int):
      return Const(int(expr.value))
    else:
      return unknown
  
  def visit_Attribute(self, expr):
    value = self.visit_expr(expr.value)
    if value.__class__ is Tuple:
      # assume tuples lowered into structs at this point
      pos = expr.value.type.field_pos(expr.name)
      return value.elts[pos]
    elif value.__class__ is Array and expr.name == 'strides':
      return value.strides
    elif value.__class__ is Struct and expr.name in value.fields:
      return value.fields[expr.name]
    else:
      return unknown

  def visit_TupleProj(self, expr):
    value = self.visit_expr(expr.tuple)
    if isinstance(value, Tuple):
      return value.elts[expr.index]
    else:
      return unknown 
    
  def visit_PrimCall(self, expr):
    abstract_values = self.visit_expr_list(expr.args)
    if all(v.__class__ is Const for v in abstract_values):
      ints = [v.value for v in abstract_values]
      return Const(expr.prim.fn(*ints))
    else:
      return unknown

  def visit_Struct(self, expr):
    if expr.type.__class__ is TupleT:
      return abstract_tuple(self.visit_expr_list(expr.args))
    elif expr.type.__class__ is ArrayT:
      stride_pos = expr.type.field_pos("strides")
      stride_arg = expr.args[stride_pos]
      stride_val = self.visit_expr(stride_arg)
      return Array(stride_val)
    else:
      return unknown   
    
  def visit_Alloc(self, expr):
    return unknown 
  
  def visit_Index(self, expr):
    return unknown 
  
  def visit_Assign(self, stmt):
    if stmt.lhs.__class__ is Var:
      value = self.visit_expr(stmt.rhs)
      assert value is not None, \
        "%s : %s returned None in ConstantStride analysis" % (stmt.rhs, stmt.rhs.node_type()) 
      self.env[stmt.lhs.name] = value 
      
  def visit_Return(self, stmt):
    value = self.visit_expr(stmt.value)
    if self.return_value is None:
      self.return_value = value 
    elif self.return_value != value:
      self.return_value = unknown 


import numpy as np 
from compare_perf import compare_perf

# Simple convolution of 3x3 patches from a given array x
# by a 3x3 array of filter weights
 
def conv_3x3_trim(x, weights):
  return np.array([[(x[i-1:i+2, j-1:j+2]*weights).sum() 
                    for j in xrange(1, x.shape[1] -2)]
                    for i in xrange(1, x.shape[0] -2)])
 

x = np.random.randn(1200,1200).astype('float32')
w = np.random.randn(3,3).astype('float32')
compare_perf(conv_3x3_trim, [x,w])


w = np.random.randn(3,3).astype('float32')
# Simple convolution of 5x5 patches from a given array x
# by a 5x5 array of filter weights
 
def conv_3x3_trim_loops(image, weights):
  result = np.zeros_like(image)
  for i in xrange(1,x.shape[0]-1):
    for j in xrange(1,x.shape[1]-1):
      for ii in xrange(3): 
        for jj in xrange(3):
          result[i,j] += image[i-ii+1, j-jj+1] * weights[ii, jj] 
  return result


compare_perf(conv_3x3_trim_loops, [x,w])

import parakeet 
def conv_3x3_imap(image, weights):
  def compute((i,j)):
      total = np.float32(0.0)
      for ii in xrange(3):
          for jj in xrange(3):
            total += image[i+ii-1, j + jj - 1] * weights[ii, jj]
      return total 
  w,h = image.shape
  return parakeet.imap(compute, (w-2,h-2))
compare_perf(conv_3x3_imap, [x,w])



import collections 
import imp 
import os

from tempfile import NamedTemporaryFile

from .. import config as root_config 
import config 
from system_info import (python_lib_dir,  
                         windows,  
                         get_source_extension, object_extension, shared_extension,  
                         get_compiler, 
                         include_dirs)
from flags import get_compiler_flags, get_linker_flags
from shell_command import CommandFailed, run_cmd 

CompiledPyFn = collections.namedtuple("CompiledPyFn",
                                      ("c_fn", 
                                       "src_filename", 
                                       "module", 
                                       "shared_filename", 
                                       "src", 
                                       "fn_name",
                                       "fn_signature"))

CompiledObject = collections.namedtuple("CompiledObject", 
                                        (
                                         "src_filename",  
                                         "object_filename", 
                                         "fn_name",
                                         ))

  
c_headers = ["stdint.h",  "math.h",  "signal.h"]
core_python_headers = ["Python.h"]
numpy_headers = ['numpy/arrayobject.h', 'numpy/arrayscalars.h']

python_headers = core_python_headers + numpy_headers 

# went to some annoying effort to clean up all the array->flags, &c that have been 
# replaced with PyArray_FLAGS in NumPy 1.7 
global_preprocessor_defs = ["#define NPY_NO_DEPRECATED_API NPY_1_7_API_VERSION"]


def create_module_source(raw_src, fn_name, 
                            extra_headers = [], 
                            declarations = [], 
                            extra_function_sources = [], 
                            print_source = None):
    # when compiling with NVCC, other headers get implicitly included 
  # and cause warnings since Python redefines this constant
  src_lines = list(global_preprocessor_defs) 
  if config.undef_posix_c_source:
    src_lines.append("#undef _XOPEN_SOURCE")
    src_lines.append("#undef _POSIX_C_SOURCE")
  
  for header in extra_headers + c_headers:
    src_lines.append("#include <%s>" % header)
  
  for decl in declarations:
    decl = decl.strip()
    if not decl.endswith(";"):
      decl += ";"
    src_lines.append(decl)
    
  src_lines.extend(extra_function_sources)
  
  src_lines.append(raw_src)
  module_init = """
    \n\n
    static PyMethodDef %(fn_name)sMethods[] = {
      {"%(fn_name)s",  %(fn_name)s, METH_VARARGS,
       "%(fn_name)s"},

      {NULL, NULL, 0, NULL}        /* Sentinel */
    };
  
    PyMODINIT_FUNC
    init%(fn_name)s(void)
    {
      //Py_Initialize();
      Py_InitModule("%(fn_name)s", %(fn_name)sMethods);
      import_array();
    }
  """ % locals()
  src_lines.append(module_init)
  full_src =  "\n".join(src_lines)
  
  if print_source is None: print_source = root_config.print_generated_code  
  if print_source:
    for i, line in enumerate(full_src.splitlines()):
        if config.print_line_numbers: print i+1, " ", line
        else: print line
  return full_src    

def create_source_file(src, 
                         fn_name = None, 
                         src_filename = None, 
                         src_extension = None):
  if fn_name is None: prefix = "parakeet_"
  else: prefix = "parakeet_%s_" % fn_name
  if src_extension is None: src_extension = get_source_extension()
  if src_filename is None:
    src_file = NamedTemporaryFile(suffix = src_extension,  
                                  prefix =  prefix, 
                                  delete = False,
                                  mode = 'w', 
                                  )
    src_filename = src_file.name 
  else:
    src_file = open(src_filename, 'w')
  
  src_file.write(src)  
  src_file.close()
  return src_file 

def compile_object(src_filename, 
                     fn_name = None,  
                     src_extension = None, 
                     declarations = [],
                     extra_objects = [], 
                     extra_compile_flags = [], 
                     print_commands = None, 
                     compiler = None, 
                     compiler_flag_prefix = None):
  
  
  if print_commands is None:  print_commands = config.print_commands
  if src_extension is None: src_extension = get_source_extension()
  if compiler is None: compiler = get_compiler()
    
  object_name = src_filename.replace(src_extension, object_extension)
  compiler_flags = get_compiler_flags(extra_compile_flags, compiler_flag_prefix)
  
  if isinstance(compiler, (list,tuple)):
    compiler_cmd = list(compiler)
  else:
    compiler_cmd = [compiler]
    
  compiler_cmd += compiler_flags 
  compiler_cmd += ['-c', src_filename, '-o', object_name]
  run_cmd(compiler_cmd, label = "Compile source")
  
  return CompiledObject(src_filename = src_filename, 
                        object_filename = object_name, 
                        fn_name = fn_name)

def link_module(compiler, object_name, shared_name, 
                 extra_objects = [], 
                 extra_link_flags = [], 
                 linker_flag_prefix = None):
  linker_flags = get_linker_flags(extra_link_flags, linker_flag_prefix) 
  
  if isinstance(compiler, (list,tuple)):
    linker_cmd = list(compiler)
  else:
    linker_cmd = [compiler]
  linker_cmd += [object_name] 
  linker_cmd += linker_flags 
  linker_cmd += list(extra_objects) 
  linker_cmd += ['-o', shared_name]

  env = os.environ.copy()
  if not windows:
    env["LD_LIBRARY_PATH"] = python_lib_dir
  run_cmd(linker_cmd, env = env, label = "Linking")

def compile_with_distutils(extension_name, 
                              src_filename,
                              extra_objects = [], 
                              extra_compiler_flags = [],
                              extra_link_flags = [],   
                              print_commands = False):

    # copied largely from pyxbuild 
    from distutils.dist import Distribution
    from distutils.extension import Extension
    
    compiler_flags = get_compiler_flags(extra_compiler_flags)
    linker_flags = get_linker_flags(extra_link_flags)
    
    ext = Extension(name=extension_name, 
                    sources=[src_filename],
                    include_dirs = include_dirs,  
                    extra_objects=extra_objects,
                    extra_compile_args=compiler_flags,
                    extra_link_args=linker_flags)
  
    script_args = ['build_ext', '--quiet']
    setup_args = {"script_name": None,
                  "script_args": script_args, 
                  }
    dist = Distribution(setup_args)
    if not dist.ext_modules: dist.ext_modules = []
    dist.ext_modules.append(ext)
    # I have no idea how distutils works or why I have to do any of this 
    config_files = dist.find_config_files()
    try: config_files.remove('setup.cfg')
    except ValueError: pass
    dist.parse_config_files(config_files)
    dist.parse_command_line()
    obj_build_ext = dist.get_command_obj("build_ext")
    
    dist.run_commands()
    shared_name = obj_build_ext.get_outputs()[0]
    return shared_name
  
def compiler_is_gnu(compiler):
  if isinstance(compiler, (list,tuple)): compiler = compiler[0]
  return (compiler.endswith("gcc") or
          compiler.endswith("gcc.exe") or 
          compiler.endswith("g++") or 
          compiler.endswith("g++.exe"))
  
def compile_module_from_source(
      partial_src, 
      fn_name,
      fn_signature = None,  
      src_filename = None,
      src_extension = None, 
      declarations = [],
      extra_function_sources = [], 
      extra_headers = [],  
      extra_objects = [],
      extra_compile_flags = [], 
      extra_link_flags = [], 
      print_source = None, 
      print_commands = None, 
      compiler = None, 
      compiler_flag_prefix = None, 
      linker_flag_prefix = None):
  
  if print_source is None: print_source = root_config.print_generated_code 
  if print_commands is None: print_commands = config.print_commands
  if src_extension is None: src_extension = get_source_extension()
  
  full_src = create_module_source(partial_src, fn_name, 
                                 extra_headers = python_headers + extra_headers, 
                                 declarations = declarations,  
                                 extra_function_sources = extra_function_sources, 
                                 print_source = print_source)
  
  src_file = create_source_file(full_src, 
                                fn_name = fn_name,
                                src_filename = src_filename, 
                                src_extension = src_extension)
  src_filename = src_file.name
  
  if compiler is None: compiler = get_compiler()
  
  try:
    compiled_object = compile_object(src_filename, 
                                     fn_name = fn_name,
                                     src_extension = src_extension,  
                                     extra_objects = extra_objects,
                                     extra_compile_flags = extra_compile_flags, 
                                     print_commands = print_commands, 
                                     compiler = compiler, 
                                     compiler_flag_prefix = compiler_flag_prefix)
  
    object_name = compiled_object.object_filename
    shared_name = src_filename.replace(src_extension, shared_extension)
    link_module(compiler, object_name, shared_name, 
                extra_objects = extra_objects, 
                extra_link_flags = extra_link_flags, 
                linker_flag_prefix = linker_flag_prefix)
    
    if config.delete_temp_files:
      os.remove(object_name)
  except CommandFailed:
    # if normal compilation fails, try distutils instead
    if not compiler_is_gnu(compiler):
      raise 
    if compiler_flag_prefix or linker_flag_prefix:
      raise 
     
    import hashlib
    digest = hashlib.sha224(full_src).hexdigest()
    shared_name = compile_with_distutils(fn_name + "_" + digest, 
                                         src_filename,
                                         extra_objects, 
                                         extra_compile_flags,
                                         extra_link_flags,
                                         print_commands)
    

  if print_commands:
    print "Loading newly compiled extension module %s..." % shared_name
  module =  imp.load_dynamic(fn_name, shared_name)
  
  #on a UNIX-style filesystem it should be OK to delete a file while it's open
  #since the inode will just float untethered from any name
  #If we ever support windows we should find some other way to delete the .dll 
  
  
  c_fn = getattr(module,fn_name)
  
  if config.delete_temp_files:
    os.remove(src_filename)
    # window's can't just untether inodes like a UNIX
    # ...have to eventually think of a plan to clean these things up
    if not windows: os.remove(shared_name)
    
  compiled_fn = CompiledPyFn(c_fn = c_fn, 
                             module = module, 
                             shared_filename =  shared_name,
                             src = full_src, 
                             src_filename = src_filename,
                             fn_name = fn_name, 
                             fn_signature = fn_signature)
  return compiled_fn




import numpy as np



def harris(I):
  m,n = I.shape
  dx = (I[1:, :] - I[:m-1, :])[:, 1:]
  dy = (I[:, 1:] - I[:, :n-1])[1:, :]

  #
  #   At each point we build a matrix
  #   of derivative products
  #   M =
  #   | A = dx^2     C = dx * dy |
  #   | C = dy * dx  B = dy * dy |
  #
  #   and the score at that point is:
  #      det(M) - k*trace(M)^2
  #
  A = dx * dx
  B = dy * dy
  C = dx * dy
  tr = A + B
  det = A * B - C * C
  k = 0.05
  return det - k * tr * tr

from compare_perf import compare_perf 
m,n = 1920, 1080
dtype = 'float64'
I = np.random.randn(m,n).astype(dtype)

compare_perf(harris, [I], propagate_exceptions=True, backends= ("c", "openmp"))



#!/usr/bin/env python

from setuptools import setup, find_packages
import os 

import parakeet.package_info

readme_filename = os.path.join(os.path.dirname(__file__), 'README.md')
with open(readme_filename, 'r') as f:
  readme = f.read()
  
try:
  import pypandoc
  readme = pypandoc.convert(readme, to='rst', format='md')
except:
  print "Conversion of long_description from markdown to reStructuredText failed, skipping..."

setup(
    name="parakeet",
    description="Runtime compiler for numerical Python.",
    long_description=readme,
    classifiers=['Development Status :: 3 - Alpha',
                 'Topic :: Software Development :: Libraries',
                 'License :: OSI Approved :: BSD License',
                 'Intended Audience :: Developers',
                 'Programming Language :: Python :: 2.7',
                 ],
    author="Alex Rubinsteyn",
    author_email="alexr@cs.nyu.edu",
    license="BSD",
    version=parakeet.package_info.__version__,
    url=parakeet.package_info.__website__, 
    download_url = 'https://github.com/iskandr/parakeet/releases', 
    packages=find_packages() + ['parakeet.test', 'parakeet.benchmarks', 'parakeet.examples'], 
    package_dir={ 
                  'parakeet.benchmarks' : './benchmarks', 
                  'parakeet.test' : './test',
                  'parakeet.examples' : './examples', 
                },
    install_requires=[
      'numpy>=1.7',       
      'dsltools',
      #'appdirs', 
      # LLVM is optional as long as you use the C backend 
      # 'llvmpy',
    ])



from .. import names 
from ..ndtypes import BoolT, IntT 
import type_mappings 
from reserved_names import is_reserved


class BaseCompiler(object):
  
  def __init__(self, extra_link_flags = None, extra_compile_flags = None):
    self.blocks = []
    self.name_versions = {}
    self.name_mappings = {}
    self.extra_link_flags = extra_link_flags if extra_link_flags else []
    self.extra_compile_flags = extra_compile_flags if extra_compile_flags else []
    
  def add_compile_flag(self, flag):
    if flag not in self.extra_compile_flags:
      self.extra_compile_flags.append(flag)
  
  def add_link_flag(self, flag):
    if flag not in self.extra_link_flags:
      self.extra_link_flags.append(flag)
        
  
  def visit_expr(self, expr):
    expr_class_name = expr.__class__.__name__
    method_name = "visit_" + expr_class_name
    assert hasattr(self, method_name), "Unsupported expression %s" % expr_class_name  
    result = getattr(self, method_name)(expr)
    assert result is not None, \
      "Compilation method for expression %s returned None, expected code string" % expr_class_name
    return result 
      
  def visit_expr_list(self, exprs):
    return [self.visit_expr(e) for e in exprs]
  
  def breakpoint(self):
    self.append("raise(SIGINT);")
    
  def visit_stmt(self, stmt):
    stmt_class_name = stmt.__class__.__name__
    method_name = "visit_" + stmt_class_name
    assert hasattr(self, method_name), \
      "Statement %s not supported by %s" % (stmt_class_name, self.__class__.__name__)  
    result = getattr(self, method_name)(stmt)
    assert result is not None, "Compilation method for statement %s return None" % stmt_class_name
    return result 
  
  def push(self):
    self.blocks.append([])
  
  def pop(self):
    stmts = self.blocks.pop()
    return "  " + self.indent("\n".join("  " + stmt for stmt in stmts))
  
  def indent(self, block_str):
    return block_str.replace("\n", "\n  ")
  
  def append(self, stmt):
    stripped = stmt.strip()
    
    assert len(stripped) == 0 or \
      ";" in stripped or \
      stripped.startswith("//") or \
      stripped.startswith("/*"), "Invalid statement: %s" % stmt
    self.blocks[-1].append(stmt)
  
  def newline(self):
    self.append("\n")
    
  def comment(self, text):
    self.append("// %s" % text)
  
  def printf(self, fmt, *args):
    
    result = 'printf("%s\\n"' % fmt
    if len(args) > 0:
      result = result + ", " + ", ".join(str(arg) for arg in args)
    self.append( result + ");" )
  
  def fresh_name(self, prefix):
    prefix = names.original(prefix)
    
    prefix = prefix.replace(".", "")
  
    version = self.name_versions.get(prefix, 1)
    self.name_versions[prefix] = version + 1
    
    # not valid chars!
    if not any(c.isalpha() for c in prefix):
      prefix = self.fresh_name("temp" + prefix) 
      
    if version == 1 and not is_reserved(prefix):
      return prefix 
    elif prefix[-1] != "_":
      return "%s_%d" % (prefix, version)
    else:
      return prefix + str(version)
    
  def to_ctype(self, t):
    """
    Convert Parakeet type to string representing its C type.
    The base class implementation only handles scalars, 
    support for Tuples, Slices, and Arrays is in the overload FlatFnCompiler.to_ctype
    """
    return type_mappings.to_ctype(t)
  
  def fresh_var(self, t, prefix = None, init = None):
    if prefix is None:
      prefix = "temp"
    name = self.fresh_name(prefix)
    if isinstance(t, str):
      t_str = t
    else:
      t_str = self.to_ctype(t)
    if init is None:
      self.append("%s %s;" % (t_str, name))
    else:
      self.append("%s %s = %s;" % (t_str, name, init))
    return name
  
  def fresh_array_var(self, t, n, prefix = None):
    if prefix is None:
      prefix = "temp"
    name = self.fresh_name(prefix)
    if isinstance(t, str):
      t_str = t
    else:
      t_str = self.to_ctype(t)
    self.append("%s %s[%d];" % (t_str, name, n))
    return name
  
  def assign(self, name, rhs):
    self.append("%s = %s;" % (name, rhs))
  
  def name(self, ssa_name, overwrite = False):
    """
    Convert from ssa names, which might have large version numbers and contain 
    syntactically invalid characters to valid local C names
    """
    if ssa_name in self.name_mappings and not overwrite:
      return self.name_mappings[ssa_name]
    prefix = names.original(ssa_name)
    prefix = prefix.replace(".", "")

    name = self.fresh_name(prefix)
    self.name_mappings[ssa_name] = name 
    return name
   
  def return_if_null(self, obj):
    self.append("if (!%s) { return NULL; }" % obj)
    
  def not_(self, x):
    if x == "1":
      return "0"
    elif x == "0":
      return "1"
    return "!%s" % x
  
  def and_(self, x, y):
    if x == "0" or y == "0":
      return "0"
    elif x == "1" and y == "1":
      return "1"
    elif x == "1":
      return y 
    elif y == "1":
      return x
    return "%s && %s" % (x,y) 
  
  def or_(self, x, y):
    if x == "1" or y == "1":
      return "1"
    elif x == "0":
      return y
    elif y == "0":
      return x 
    return "%s || %s" % (x,y) 
  
  def gt(self, x, y, t):
    if isinstance(t, (BoolT, IntT)) and x == y:
      return "0"
    return "%s > %s" % (x, y)
  
  def gte(self, x, y, t):
    if isinstance(t, (BoolT, IntT)) and x == y:
      return "1"
    return "%s >= %s" % (x,y) 
  
  def lt(self, x, y, t):
    if isinstance(t, (BoolT, IntT)) and x == y:
      return "0"
    return "%s < %s" % (x,y)
  
  def lte(self, x, y, t):
    if isinstance(t, (BoolT, IntT)) and x == y:
      return "1"
    return "%s <= %s" % (x, y) 
  
  def neq(self, x, y, t):
    if isinstance(t, (BoolT, IntT)) and x == y:
      return "0"
    return "%s != %s" % (x, y) 
  
  def eq(self, x, y, t):
    if isinstance(t, (BoolT, IntT)) and x == y:
      return "1"
    return "%s == %s" % (x, y)
  
  
  def add(self, x, y):
    if x == "0":
      return y
    elif y == "0":
      return x
    return "%s + %s" % (x,y)
  
  def sub(self, x, y):
    if x == "0":
      return "-(%s)" % y 
    elif y == "0":
      return x 
    return "%s - %s" % (x,y)
  
  def mul(self, x, y):
    if x == "1":
      return y 
    elif y == "1":
      return x 
    elif x == 0 or y == 0:
      return "0"
    return "%s * %s" % (x,y)
  
  def div(self, x, y):
    if x == y:
      return "1"
    elif x == "0":
      return "0"
    elif y == "1":
      return x
    else:
      return "%s / %s" % (x,y)
    


import numpy as np 
from parakeet import testing_helpers




mu = 0.1
Lx, Ly = 7, 7

def diffuse_loops(iter_num):
    u = np.zeros((Lx, Ly), dtype=np.float64)
    temp_u = np.zeros_like(u)
    temp_u[Lx / 2, Ly / 2] = 1000.0
    for _ in range(iter_num):
        for i in range(1, Lx - 1):
            for j in range(1, Ly - 1):
                u[i, j] = mu * (temp_u[i + 1, j] + temp_u[i - 1, j] +
                                temp_u[i, j + 1] + temp_u[i, j - 1] -
                                4 * temp_u[i, j])
        temp = u
        u = temp_u
        temp_u = temp
    return u

def test_diffuse_loops():
  testing_helpers.expect(diffuse_loops, [3], diffuse_loops(3))


def diffuse_array_expressions(iter_num):
    u = np.zeros((Lx, Ly), dtype=np.float64)
    temp_u = np.zeros_like(u)
    temp_u[Lx / 2, Ly / 2] = 1000.0

    for _ in range(iter_num):
        u[1:-1, 1:-1] = mu * (temp_u[2:, 1:-1] + temp_u[:-2, 1:-1] +
                              temp_u[1:-1, 2:] + temp_u[1:-1, :-2] -
                              4 * temp_u[1:-1, 1:-1])
        temp = u
        u = temp_u
        temp_u = temp
    return u

def test_diffuse_array_expressions():
  testing_helpers.expect(diffuse_array_expressions, [2], diffuse_array_expressions(2))

if __name__ == "__main__":
  testing_helpers.run_local_tests()




from parakeet.testing_helpers import expect, run_local_tests

def test_max_int():
  expect(max, [3,4], 4)

def test_max_float():
  expect(max, [3.0,4.0], 4.0)
  
def test_max_int_float():
  expect(max, [3, 4.0], 4.0)

def test_max_bool():
  expect(max, [False, True], True)
  
if __name__ == '__main__':
  run_local_tests()



from seq_expr import SeqExpr 

class List(SeqExpr):
  def __init__(self, elts, type = None, source_info = None):
    self.elts = tuple(elts)
    self.type = type 
    self.source_info = source_info
  
  def children(self):
    return self.elts 
  


from transform import Transform

class IndexMapElimination(Transform):
  """
  a = IndexMap(f, bounds)
  ... a[i] ... 
  
  -becomes-
  
  ... f(i) ... 
  """
  def transform_IndexMap(self, expr):
    return expr 


from expr import Expr 
from stmt import Stmt 

class Struct(Expr):
  """
  Eventually all non-scalar data should be transformed to be created with this
  syntax node, signifying explicit struct allocation
  """

  def __init__(self, args, type = None, source_info = None):
    self.args = tuple(args)
    self.type = type 
    self.source_info = source_info 
    
  def __str__(self):
    return "Struct(%s) : %s" % \
           (", ".join(str(arg) for arg in self.args), self.type)

  def children(self):
    return self.args

  def __hash__(self):
    self.args = tuple(self.args)
    return hash(self.args)

class Alloc(Expr):
  """Allocates a block of data, returns a pointer"""
  
  def __init__(self, elt_type, count, type = None, source_info = None):
    self.elt_type = elt_type 
    self.count = count 
    self.type = type 
    self.source_info = source_info
  
  def __str__(self):
    return "alloc<%s>[%s] : %s" % (self.elt_type, self.count, self.type)

  def children(self):
    return (self.count,)

  def __hash__(self):
    return hash((self.elt_type, self.count))

class Free(Expr):
  """Free a manually allocated block of memory"""
  def __init__(self, value, type = None, source_info = None):
    self.value = value 
    self.type = type 
    self.source_info = source_info
    
  def __str__(self):
    return "free(%s)" % self.value 
  
  def children(self):
    yield self.value 
    
  def __hash__(self):
    return hash(self.value)
  
class NumCores(Expr):
  
  """
  Degree of available parallelism, 
  varies depending on backend and how
  ParFor is actually being mapped 
  to executing threads/thread blocks/etc..
  """
  
  def __str__(self):
    return "NUM_CORES"
  
  def __eq__(self, other):
    return other.__class__ is NumCores 
  
  def node_init(self):
    from ..ndtypes import Int64
    self.type = Int64 
  
  def __hash__(self):
    return 0
  
  def children(self):
    return ()

class SourceExpr(Expr):
  """
  Splice this code directly into the low-level representation, 
  should only be used from within a backend that knows what
  the target code should look like 
  """
  def __init__(self, text, type = None, source_info = None):
    self.text = text 
    self.type = type 
    self.source_info = source_info 
  
  def __str__(self):
    return "SourceExpr(%s)" % self.text 

class SourceStmt(Stmt):
  """
  Splice this code directly into the low-level representation, 
  should only be used from within a backend that knows what
  the target code should look like 
  """
  def __init__(self, text, type = None, source_info = None):
    self.text = text 
    self.type = type 
    self.source_info = source_info 
    
  def __str__(self):
    return "SourceStmt(%s)" % self.text 
  